from __future__ import print_function

import sys
from optparse import OptionParser

import numpy
from matplotlib import pylab as plt

from glue.lal import Cache
import lalsimutils

"""
test_like_and_samp.py:  Testing the likelihood evaluation and sampler, working in conjunction

"""

__author__ = "Evan Ochsner <evano@gravity.phys.uwm.edu>, Chris Pankow <pankow@gravity.phys.uwm.edu>, R. O'Shaughnessy"

from factored_likelihood import *
import ourio

checkInputs = False

rosUseDifferentWaveformLengths = False    
rosUseRandomTemplateStartingFrequency = False

rosUseTargetedDistance = True
rosUseStrongPriorOnParameters = False
rosDebugMessages = False
rosShowSamplerInputDistributions = True
rosShowRunningConvergencePlots = True
rosShowTerminalSampleHistograms = True
rosSaveHighLikelihoodPoints = True
rosUseThresholdForReturn = True
rosUseMultiprocessing= False
rosDebugCheckPriorIntegral = False
nMaxEvals = 1e4

if rosUseThresholdForReturn and nMaxEvals > 5e4:
    fracThreshold = 0.95
else:
    fracThreshold = 0.0


theEpochFiducial = lal.LIGOTimeGPS(1064023405.000000000)   # 2013-09-24 early am 
tEventFiducial = 0                                                                 # relative to GPS reference

optp = OptionParser()
optp.add_option("-c", "--cache-file", default=None, help="LIGO cache file containing all data needed.")
optp.add_option("-C", "--channel-name", action="append", help="instrument=channel-name, e.g. H1=FAKE-STRAIN. Can be given multiple times for different instruments.")
opts, args = optp.parse_args()

det_dict = {}
rhoExpected ={}
if opts.channel_name is not None and opts.cache_file is None:
    print("Cache file required when requesting channel data.", file=sys.stderr)
    exit(-1)
elif opts.channel_name is not None:
    det_dict = dict(map(lambda cname: cname.split("="), opts.channel_name))

Niter = 5 # Number of times to call likelihood function
Tmax = 0.05 # max ref. time
Tmin = -0.05 # min ref. time
Dmax = 110. * 1.e6 * lal.LAL_PC_SI # max distance
Dmin = 1. * 1.e6 * lal.LAL_PC_SI   # min distance

ampO =0 # sets which modes to include in the physical signal
Lmax = 2 # sets which modes to include
fref = 100
fminWavesSignal = 25  # too long can be a memory and time hog, particularly at 16 kHz
fSample = 4096*4

if rosUseDifferentWaveformLengths: 
    fminWavesTemplate = fminWavesSignal+0.005
else:
    if rosUseRandomTemplateStartingFrequency:
         print("   --- Generating a random template starting frequency  ---- ")
         fminWavesTemplate = fminWavesSignal+5.*np.random.random_sample()
    else:
        fminWavesTemplate = fminWavesSignal

#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
if len(det_dict) > 0:
    for d, chan in det_dict.items():
        data_dict[d] = lalsimutils.frame_data_to_hoff(opts.cache_file, chan)
else:
    m1 = 4*lal.LAL_MSUN_SI
    m2 = 3*lal.LAL_MSUN_SI

    Psig = ChooseWaveformParams(
        m1 = m1,m2 =m2,
        fmin = fminWavesSignal, 
        fref=fref, ampO=ampO,
                                radec=True, theta=1.2, phi=2.4,
                                detector='H1', 
                                dist=25.*1.e6*lal.LAL_PC_SI,
                                tref = theEpochFiducial,
        deltaT=1./fSample
                                )
    df = findDeltaF(Psig)
    Psig.print_params()
    Psig.deltaF = df
    data_dict['H1'] = lalsimutils.non_herm_hoff(Psig)
    Psig.detector = 'L1'
    data_dict['L1'] = lalsimutils.non_herm_hoff(Psig)
    Psig.detector = 'V1'
    data_dict['V1'] = lalsimutils.non_herm_hoff(Psig)

# TODO: Read PSD from XML
psd_dict = {}
analyticPSD_Q = True # For simplicity, using an analytic PSD
psd_dict['H1'] = lal.LIGOIPsd
psd_dict['L1'] = lal.LIGOIPsd
psd_dict['V1'] = lal.LIGOIPsd

print(" == Data report == ")
detectors = data_dict.keys()
rho2Net = 0
print(" Amplitude report :")
fminSNR =30
for det in detectors:
    IP = ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=Psig.deltaF,psd=psd_dict[det])
    rhoExpected[det] = rhoDet = IP.norm(data_dict[det])
    rho2Net += rhoDet*rhoDet
    print(det, " rho = ", rhoDet)
print("Network : ", np.sqrt(rho2Net))

if checkInputs:
    print(" == Plotting detector data (time domain; requires regeneration, MANUAL TIMESHIFTS,  and seperate code path! Argh!) == ")
    P = Psig.copy()
    P.tref = Psig.tref
    for det in detectors:
        P.detector=det   # we do 
        hT = hoft(P)
        tvals = float(P.tref - theEpochFiducial) + hT.deltaT*np.arange(len(hT.data.data))
        plt.figure(1)
        plt.plot(tvals, hT.data.data,label=det)

    tlen = hT.deltaT*len(np.nonzero(np.abs(hT.data.data)))
    tRef = np.abs(float(hT.epoch))
    plt.xlim( tRef-0.5,tRef+0.1)  # Not well centered, based on epoch to identify physical merger time
    plt.savefig("test_like_and_samp-input-hoft.pdf")



# Struct to hold template parameters
P = ChooseWaveformParams(fmin=fminWavesTemplate, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         fref=fref,
         tref=theEpochFiducial,
         deltaT=1./fSample,
         dist=100*1.e6*lal.LAL_PC_SI,
         deltaF=df)


#
# Perform the Precompute stage
#
rholms_intp, crossTerms, rholms, epoch_post = PrecomputeLikelihoodTerms(theEpochFiducial,P, data_dict,psd_dict, Lmax, analyticPSD_Q)
print("Finished Precomputation...")
print("====Generating metadata from precomputed results =====")
distBoundGuess = estimateUpperDistanceBoundInMpc(rholms, crossTerms)
print(" distance probably less than ", distBoundGuess, " Mpc")

print("====Loading metadata from previous runs (if any): sampler-seed-data.dat =====")


if checkInputs == True:

    print(" ======= UV test: Recover the SNR of the injection  ==========")
    print(" Detector lnLmodel  (-2lnLmodel)^(1/2)  rho(directly)  [last two entries should be equal!] ")
    for det in detectors:
        lnLModel = SingleDetectorLogLikelihoodModel(crossTerms, Psig.tref, Psig.phi, Psig.theta, Psig.incl, Psig.phiref, Psig.psi, Psig.dist, 2, det)
        print(det, lnLModel, np.sqrt(-2*lnLModel), rhoExpected[det], "      [last two equal?]")
    print(" ======= End to end LogL: Recover the SNR of the injection at the injection parameters  ==========")
    lnL = FactoredLogLikelihood(Psig, rholms_intp, crossTerms, Lmax)
    print("  : Evan's code : ", lnL, " versus rho^2/2 ", rho2Net/2)
    print("  : Timing issues (checkme!) : fiducial = ", stringGPSNice(theEpochFiducial))

    print(" ======= rholm test: Plot the lnLdata timeseries at the injection parameters (* STILL TIME OFFSET *)  ==========")
    tmin = np.max(float(epoch_post - theEpochFiducial),tWindowReference[0]+0.03)   # the minimum time used is set by the rolling condition
#    tvals = np.linspace(tmin,tWindowReference[1],4*fSample*(tWindowReference[1]-tmin))
    tvals = np.linspace(tWindowExplore[0]+tEventFiducial,tWindowExplore[1]+tEventFiducial,fSample*(tWindowExplore[1]-tWindowExplore[0]))
    for det in detectors:
        lnLData = map( lambda x: SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, theEpochFiducial+x, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, 2, det), tvals)
        lnLDataEstimate = np.ones(len(tvals))*rhoExpected[det]*rhoExpected[det]
        plt.figure(1)
        tvalsPlot = tvals 
        plt.plot(tvalsPlot, lnLData,label='Ldata(t)+'+det)
        plt.plot(tvalsPlot, lnLDataEstimate,label="$rho^2("+det+")$")
    tEventRelative =float( Psig.tref - theEpochFiducial)
    print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is ", tEventRelative)
    plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
    plt.title("lnLdata (interpolated) vs narrow time interval")
    plt.xlabel('t(s)')
    plt.ylabel('lnLdata')

    print(" ======= rholm test: Plot the lnL timeseries at the injection parameters (* STILL TIME)  ==========")
    tmin = np.max(float(epoch_post - theEpochFiducial),tWindowReference[0]+0.03)   # the minimum time used is set by the rolling condition
    tvals = np.linspace(tmin,tWindowReference[1],4*fSample*(tWindowReference[1]-tmin))
    P = Psig.copy()
    lnL = np.zeros(len(tvals))
    for indx in np.arange(len(tvals)):
            P.tref =  theEpochFiducial+tvals[indx]
            lnL[indx] =  FactoredLogLikelihood(P, rholms_intp, crossTerms, 2)
    lnLEstimate = np.ones(len(tvals))*rho2Net/2
    plt.figure(1)
    tvalsPlot = tvals 
    plt.plot(tvalsPlot, lnL,label='lnL(t)')
    plt.plot(tvalsPlot, lnLEstimate,label="$rho^2/2(net)$")
    tEventRelative =float( Psig.tref - theEpochFiducial)
    print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is the same ", tEventRelative)
    plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
    plt.title("lnL (interpolated) vs narrow time interval")
    plt.xlim(Tmin,Tmax)  # the window we actually use
    plt.legend()
    plt.savefig("test_like_and_samp-lnLvsTime.pdf")


#
# Call the likelihood function for various extrinsic parameter values
#
nEvals = 0
def likelihood_function(phi, theta, tref, phiref, incl, psi, dist):
    global nEvals
    global pdfFullPrior

    lnL = numpy.zeros(phi.shape)
    i = 0
#    print " Likelihood results :  "
#    print " iteration Neff  lnL   sqrt(2max(lnL))  rho  sqrt(2 lnLmarg)   <lnL> "
    for ph, th, tr, phr, ic, ps, di in zip(phi, theta, tref, phiref, incl, psi, dist):
        P.phi = ph # right ascension
        P.theta = th # declination
        P.tref = theEpochFiducial + tr # ref. time (rel to epoch for data taking)
        P.phiref = phr # ref. orbital phase
        P.incl = ic # inclination
        P.psi = ps # polarization angle
        P.dist = di # luminosity distance

        lnL[i] = NetworkLogLikelihoodPolarizationMarginalized(theEpochFiducial, rholms_intp, crossTerms,Psig.tref, Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist, 2, detectors)
        i+=1


    nEvals+=i 
    return numpy.exp(lnL)

import RIFT.integrators.mcsampler as mcsampler
sampler = mcsampler.MCSampler()

# Sampling distribution
def uniform_samp(a, b, x):
   if type(x) is float:
       return 1/(b-a)
   else:
       return numpy.ones(x.shape[0])/(b-a)

# set up bounds on parameters
# Polarization angle
psi_min, psi_max = 0, numpy.pi
# RA and dec
ra_min, ra_max = 0, 2*numpy.pi
dec_min, dec_max = -numpy.pi/2, numpy.pi/2
# Reference time
tref_min, tref_max = Tmin, Tmax
# Inclination angle
inc_min, inc_max = 0, numpy.pi
# orbital phi
phi_min, phi_max = 0, 2*numpy.pi
# distance
dist_min, dist_max = Dmin, Dmax

import functools
# Define the true prior PDF
# The likelihood function assumes this function exists.
# The integrator actually does a monte carlo integral of L*p using a sampling prior p_s.
def pdfFullPrior(phi, theta, tref, phiref, incl, psi, dist): # remember theta is dec, 
    return (np.cos(theta)/(4*np.pi)) * 1./(tWindowExplore[1]-tWindowExplore[0])*1/(3*Dmax**3)*np.cos(incl)/(4*np.pi)*1./(np.pi)


# Uniform sampling (in area) but nonuniform sampling in distance (*I hope*).  Auto-cdf inverse
if rosUseStrongPriorOnParameters:
    sampler.add_parameter("psi", functools.partial(mcsampler.gauss_samp, Psig.psi, 0.5), None, psi_min, psi_max, 
                          prior_pdf =mcsampler.uniform_samp_psi  )
    # Use few degree square prior : at SNR 20 in 3 detetors.  This is about 10 deg square
    sampler.add_parameter("ra", functools.partial(mcsampler.gauss_samp, Psig.phi,0.05), None, ra_min, ra_max, 
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("dec", functools.partial(mcsampler.gauss_samp, Psig.theta,0.05), None, dec_min, dec_max, 
                          prior_pdf= mcsampler.uniform_samp_dec)
    sampler.add_parameter("tref", functools.partial(mcsampler.gauss_samp, tEventFiducial, 0.01), None, tref_min, tref_max, 
                          prior_pdf = functools.partial(mcsampler.uniform_samp_vector, tWindowExplore[0],tWindowExplore[1]))
    sampler.add_parameter("phi", functools.partial(mcsampler.gauss_samp, Psig.phiref,0.5), None, phi_min, phi_max, 
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("incl", functools.partial(mcsampler.gauss_samp, Psig.incl,0.3), None, inc_min, inc_max, 
                          prior_pdf = mcsampler.uniform_samp_theta)
    if rosUseTargetedDistance:
        sampler.add_parameter("dist", functools.partial(mcsampler.quadratic_samp_vector,  distBoundGuess*1e6*lal.LAL_PC_SI ), None, dist_min, dist_max, prior_pdf = np.vectorize(lambda x: x**2/(3*Dmax**3)))
    else:
        sampler.add_parameter("dist", functools.partial(mcsampler.gauss_samp_withfloor, Psig.dist, Psig.dist*0.1, 0.001/Psig.dist), None, dist_min, dist_max,prior_pdf =np.vectorize( lambda x: x**2/(3*Dmax**3)))
else:
    # PROBLEM: Underlying prior samplers are not uniform.  We need two stages
    sampler.add_parameter("psi", functools.partial(mcsampler.uniform_samp_vector, psi_min, psi_max), None, psi_min, psi_max,
                          prior_pdf =mcsampler.uniform_samp_psi )
    # Use few degree square prior : at SNR 20 in 3 detetors.  This is about 10 deg square
    sampler.add_parameter("ra", functools.partial(mcsampler.gauss_samp, Psig.phi,0.05), None, ra_min, ra_max, 
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("dec", functools.partial(mcsampler.gauss_samp, Psig.theta,0.05), None, dec_min, dec_max, 
                          prior_pdf= mcsampler.uniform_samp_dec)
    sampler.add_parameter("tref", functools.partial(mcsampler.gauss_samp, tEventFiducial, 0.01), None, tref_min, tref_max, 
                          prior_pdf = functools.partial(mcsampler.uniform_samp_vector, tWindowExplore[0],tWindowExplore[1]))
    sampler.add_parameter("phi", functools.partial(mcsampler.uniform_samp_vector, phi_min, phi_max), None, phi_min, phi_max,
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("incl", functools.partial(mcsampler.cos_samp_vector), None, inc_min, inc_max,
                          prior_pdf = mcsampler.uniform_samp_theta)
    if rosUseTargetedDistance:
        sampler.add_parameter("dist", functools.partial(mcsampler.quadratic_samp_vector,  distBoundGuess*1e6*lal.LAL_PC_SI ), None, dist_min, dist_max, prior_pdf = np.vectorize(lambda x: x**2/(3*Dmax**3)))
    else:
        sampler.add_parameter("dist", functools.partial(mcsampler.gauss_samp_withfloor, Psig.dist, Psig.dist*0.1, 0.001/Psig.dist), None, dist_min, dist_max,prior_pdf =np.vectorize( lambda x: x**2/(3*Dmax**3)))


if rosShowSamplerInputDistributions:
    print(" ====== Plotting prior and sampling distributions ==== ")
    # Done by hand so I have access to them later
    nFig = 0
    pltAssociation = {}
    for param in sampler.params:
        nFig+=1
        plt.figure(nFig)
        pltAssociation[param] = nFig
        plt.clf()
        xLow = sampler.llim[param]
        xHigh = sampler.rlim[param]
        xvals = np.linspace(xLow,xHigh,500)
        pdfPrior = sampler.prior_pdf[param]  # Force type conversion in case we have non-float limits for some reasona
        pdfvalsPrior = np.array(map(pdfPrior, xvals))  # do all the numpy operations by hand: no vectorization
        pdf = sampler.pdf[param]
        cdf = sampler.cdf[param]
        pdfvals = pdf(xvals)
        cdfvals = cdf(xvals)
        if str(param) ==  "dist":
            xvvals = xvals/(1e6*lal.LAL_PC_SI)       # plot in Mpc, not m.  Note PDF has to change
            pdfvalsPrior = pdfvalsPrior * (1e6*lal.LAL_PC_SI) # rescale units
            pdfvals = pdfvals * (1e6*lal.LAL_PC_SI) # rescale units
        plt.plot(xvals,pdfvalsPrior,label="prior:"+str(param),linestyle='--')
        plt.plot(xvals,pdfvals,label=str(param))
        plt.plot(xvals,cdfvals,label='cdf:'+str(param))
        plt.xlabel(str(param))
        plt.legend()
        plt.savefig("test_like_and_samp-"+str(param)+".pdf")
#    plt.show()

tGPSStart = lal.GPSTimeNow()
res, var, ret, lnLmarg, neff = sampler.integrate(likelihood_function, "ra", "dec", "tref", "phi", "incl", "psi", "dist", n=100,nmax=nMaxEvals,igrandmax=rho2Net/2,full_output=True,neff=100,igrand_threshold_fraction=fracThreshold,use_multiprocessing=rosUseMultiprocessing,verbose=True)
tGPSEnd = lal.GPSTimeNow()
print(" Evaluation time  = ", float(tGPSEnd - tGPSStart), " seconds")
print(" lnLmarg is ", np.log(res), " with nominal sampling error ", np.sqrt(var)/res, " but a more reasonable estimate based on the lnL history is ", np.std(lnLmarg - np.log(res)))
print(" expected largest value is ", rho2Net/2, "and observed largest lnL is ", np.max(np.transpose(ret)[-1]))
print(" note neff is ", neff, "; compare neff^(-1/2) = ", 1/np.sqrt(neff))

# Save the sampled points to a file
# Only store some
ourio.dumpSamplesToFile("test_like_and_samp_margPsi-dump.dat", ret, ['ra','dec', 'tref', 'phi', 'incl', 'psi', 'dist', 'lnL'])

if checkInputs:
    plt.show()

# Plot terminal histograms from the sampled points and log likelihoods
if rosShowTerminalSampleHistograms:
    ourio.plotParameterDistributionsFromSamples("results", sampler, ret,  ['ra','dec', 'tref', 'phi', 'incl', 'psi', 'dist', 'lnL'])
    ra,dec,tref,phi,incl, psi,dist,lnL = np.transpose(ret)  # unpack. This can include all or some of the data set. The default configuration returns *all* points
    print(" ==== CONVERGENCE PLOTS (**beware: potentially truncated data!**) === ")
    plt.figure(0) # lnL
    plt.clf()
    plt.plot(np.arange(len(lnLmarg)), lnLmarg,label="lnLmarg")
    plt.plot(np.arange(len(lnL)), np.ones(len(lnL))*rho2Net/2,label="rho^2/2")
    plt.xlim(0,len(lnL))
    plt.xlabel('iteration')
    plt.ylabel('lnL')
    plt.legend()
    print(" ==== TERMINAL 1D HISTOGRAMS: Sampling and posterior === ")
    # d
    plt.figure(pltAssociation['dist']) # d = fig
    plt.clf()
    hist, bins  = np.histogram(dist/(1e6*lal.LAL_PC_SI),bins=50,density=True)
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="dist:sampled")
    hist,bins = np.histogram(dist/(1e6*lal.LAL_PC_SI),bins=50,weights=np.exp(lnL),density=True)
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="dist:post")
    # param = 'dist'
    # pdfPrior = sampler.prior_pdf[param]  # Force type conversion in case we have non-float limits for some reasona
    # pdfvalsPrior = np.array(map(pdfPrior, xvals))  # do all the numpy operations by hand: no vectorization
    # pdf = sampler.pdf[param]
    # xvvals = xvals/(1e6*lal.LAL_PC_SI)       # plot in Mpc, not m.  Note PDF has to change
    # pdfvalsPrior = pdfvalsPrior * (1e6*lal.LAL_PC_SI) # rescale units
    # pdfvals = pdfvals * (1e6*lal.LAL_PC_SI) # rescale units
    # plt.plot(xvals,pdfvalsPrior,label="prior:"+str(param),linestyle='--')
    # plt.plot(xvals,pdfvals,label=str(param))
    plt.xlabel("d (Mpc)")
    plt.title("Sampling and posterior distribution: d ")
    plt.legend()
    # tref
    plt.figure(pltAssociation['tref'])
#    plt.clf()
    hist, bins  = np.histogram(tref,bins=50,density=True)
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="tref:sampled")
    hist, bins  = np.histogram(tref,bins=50,density=True,weights=np.exp(lnL))
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="tref:post")
    plt.xlim(-0.01+tEventFiducial,0.01+tEventFiducial)
    plt.xlabel("t (s)")
    plt.title("Sampling and posterior distribution: t ")
    plt.legend()
    # incl
    plt.figure(pltAssociation['incl'])
#    plt.clf()
    hist, bins  = np.histogram(incl,bins=50,normed=True)
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="incl:sampled")
    hist, bins  = np.histogram(incl,bins=50,normed=True,weights=np.exp(lnL))
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="incl:post")
    plt.xlabel("incl")
    plt.title("Sampling and posterior distribution: incl ")
    plt.legend()
    plt.figure(4)
    plt.clf()
    hist, bins  = np.histogram(psi,bins=50,normed=True)
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="psi:sampled")
    hist, bins  = np.histogram(psi,bins=50,normed=True,weights=np.exp(lnL))
    center = (bins[:-1]+bins[1:])/2
    plt.plot(center,hist,label="psi:post")
    plt.xlabel("psi")
    plt.title("Sampling and posterior distribution: psi ")
    plt.legend()
    plt.show()
    print(" ==== TERMINAL 2D HISTOGRAMS: Sampling and posterior === ")
        # Distance-inclination
    plt.figure(1)
    plt.clf()
    H, xedges, yedges = np.histogram2d(dist/(1e6*lal.LAL_PC_SI),incl, weights=np.exp(lnL),bins=(10,10))
    extent = [yedges[0], yedges[-1], xedges[-1], xedges[0]]
    plt.imshow(H, extent=extent, interpolation='nearest', aspect=0.618)
    plt.colorbar()
    plt.xlim(0, 50)
    plt.ylim(0, np.pi)
    plt.title("Posterior distribution: d-incl ")
        # phi-psi
    plt.figure(2)
    plt.clf()
    H, xedges, yedges = np.histogram2d(phi,psi, bins=(10,10),weights=np.exp(lnL))
    extent = [yedges[0], yedges[-1], xedges[-1], xedges[0]]
    plt.imshow(H, extent=extent, interpolation='nearest', aspect=0.618)
    plt.colorbar()
    plt.xlim(0,2*np.pi)
    plt.ylim(0,np.pi)
    plt.title("Posterior distribution: phi-psi ")
        # ra-dec
    plt.figure(3)
    plt.clf()
    H, xedges, yedges = np.histogram2d(ra,dec, bins=(10,10),weights=np.exp(lnL))
    extent = [yedges[0], yedges[-1], xedges[-1], xedges[0]]
    plt.imshow(H, extent=extent, interpolation='nearest', aspect=0.618)
    plt.xlim(0,2*np.pi)
    plt.ylim(-np.pi,np.pi)
    plt.colorbar()
    plt.title("Posterior distribution: ra-dec ")
    plt.show()



#! /usr/bin/env python
#
#  EXAMPLES
#   python check_waveform_random.py --approx SpinTaylorT4 --Lmax 4
#   python check_waveform_random.py  --force-psi 0.1
#   python ./check_waveform_random.py --approx SpinTaylorT4 --force-psi 0 --use-same-fref --force-aligned
#   python check_waveform_random.py --approx SEOBNRv5EHM --force-aligned --use-eccentric --use-gwsignal --inj mdc.xml.gz --event 15
#
# RESULTS
#   - Pv2: perfect
#   - XPHM: some small phase/alignment issues.  - fixed
#   - gwsignal
#      - SEOB: perfect, now
#       - NRSur
#
# WATCH OUT FOR
#   - starting too close too isco
#   - wraparound issues (segment length)
#   - different WF interfaces: eg, some use the IMRPv2 interface fallback, some use ChooseTDModes

# NRSur: use LAL_DATA_PATH, eg from below, and use a larger total mass
#   scp ldas:/scratch/lalsimulation/NRSur7dq4_v1.0.h5 .
#    export LAL_DATA_PATH=`pwd`
#

import numpy as np
from matplotlib import pyplot as plt
import argparse
import lal
import sys

import RIFT
import RIFT.lalsimutils as lalsimutils
import RIFT.likelihood.factored_likelihood as factored_likelihood  # direct hoft call
import RIFT.physics.GWSignal as rift_gws

parser = argparse.ArgumentParser()
parser.add_argument("--approximant",type=str,default="IMRPhenomPv2")
parser.add_argument("--fiducial",action='store_true')
parser.add_argument("--mtot",default=40,type=float)
parser.add_argument("--fmin",default=20,type=float)
parser.add_argument("--use-gwsignal",action='store_true')
parser.add_argument("--use-extra-fd-args",action='store_true')
parser.add_argument("--use-xphm-spintaylor",action='store_true') # see https://git.ligo.org/asimov/data/-/blob/main/analyses/bilby-bbh/analysis_bilby_IMRPhenomXPHM-SpinTaylor.yaml?ref_type=heads
parser.add_argument("--use-same-fref",action='store_true')
parser.add_argument("--rom-group",default=None)
parser.add_argument("--rom-param",default=None)
parser.add_argument("--Lmax",default=5,type=int)
parser.add_argument("--tmin",default=-0.5,type=float)
parser.add_argument("--tmax",default=0.1,type=float)
parser.add_argument("--force-psi",default=None,type=float)
parser.add_argument("--force-phase-shift-factor-of-pi",default=0,type=float)
parser.add_argument("--force-aligned",action='store_true')
parser.add_argument("--force-zero-inclination",action='store_true')
parser.add_argument("--use-eccentric",action='store_true')
parser.add_argument("--inj", default=None,help="inspiral XML file containing injection information.")
parser.add_argument("--event",type=int, default=None,help="event ID of injection XML to use.")
parser.add_argument("--verbose",action='store_true')
opts=  parser.parse_args()

P = lalsimutils.ChooseWaveformParams()
P.ampO=-1  # need this otherwise we don't get SpinTaylor HM output
P.phaseO = 7 # so we have less insane outputs
P.taper = lalsimutils.lsu_TAPER_START
if not(opts.fiducial) and not(opts.inj):
   print("Creating random event to use for plot comparison.")
   P.randomize()
   # move inside conditional for use inj purposes
   P.dist = RIFT.likelihood.factored_likelihood.distMpcRef*1e6*lal.PC_SI  # fiducial reference distance
   P.assign_param('mtot',opts.mtot*lal.MSUN_SI)
   if opts.use_eccentric:
      P.eccentricity = np.random.uniform(0.0,0.4) #for safety, for now
      P.meanPerAno = np.random.uniform(0.0,2*np.pi)

elif opts.inj:
   if not(opts.event):
      print("ERROR: must specify event to use.")
      sys.exit(0)
   else:
      ## as in lalwriteframe
      from igwn_ligolw import lsctables, table, utils # check all are needed
      filename = opts.inj
      event = opts.event
      print(f"Using event {event} from {filename}.")
      xmldoc = utils.load_filename(filename, verbose = True, contenthandler =lalsimutils.cthdler)
      sim_inspiral_table = lsctables.SimInspiralTable.get_table(xmldoc)
      P.copy_sim_inspiral(sim_inspiral_table[int(event)])
      P.tref = 0.0 ## force this for plotting purposes only - safe?
else:
   print("Using fiducial event parameters.")
   P.m2 = P.m1/1.5
   P.theta = 0.1  # irrelevant/unused
   P.phi = 1.3    # irrelevant/unused
   P.incl = 0.2
   P.phiref = 5
   P.psi = 0.9
   P.s1x = 0.1
   P.s1y = -0.6
   P.s1z = 0.4
   P.s2x = 0.4
   P.s2z = -0.3
   if opts.use_eccentric:
      P.eccentricity = 0.15
      P.meanPerAno = np.pi
   # move inside conditional for use inj purposes
   P.dist = RIFT.likelihood.factored_likelihood.distMpcRef*1e6*lal.PC_SI  # fiducial reference distance
   P.assign_param('mtot',opts.mtot*lal.MSUN_SI)

if opts.force_aligned:
    P.s1x = P.s1y=P.s2x=P.s2y=0
if not(opts.use_gwsignal):
    if opts.approximant != "TaylorT4": # not default setting
        P.approx = lalsimutils.lalsim.GetApproximantFromString(opts.approximant)  # allow user to override the approx setting. Important for NR followup, where no approx set in sim_xml!
else:
    P.approx = opts.approximant

if opts.approximant == "SEOBNRv5EHM":
   # temp workaround - needs a solution
   P.taper = lalsimutils.lsu_TAPER_NONE
   
P.deltaT=1./4096
P.deltaF = 1./8
P.fref = 22
P.fmin=opts.fmin
if opts.use_same_fref:
    P.fref = P.fmin

if not(opts.force_psi is None):
    P.psi = opts.force_psi
if opts.force_zero_inclination:
    P.incl = 0
P.print_params()

# hoft via hlm, using exactly the function call we use in production
extra_args ={}
extra_waveform_args ={}
extra_waveform_args['fd_centering_factor']= 0.9
if opts.use_extra_fd_args:
    extra_args['fd_L_frame'] = True
if opts.use_xphm_spintaylor:
    extra_waveform_args['FinalSpinMod'] =2
    extra_waveform_args['PhenomXPHMReleaseVersion'] = 122022
    extra_waveform_args['PrecVersion'] = 320


P_copy = P.manual_copy() # beware, call may change P!
hlmF_1, _= factored_likelihood.internal_hlm_generator(P_copy, opts.Lmax, use_gwsignal=opts.use_gwsignal, use_gwsignal_approx=opts.approximant,ROM_group=opts.rom_group,ROM_param=opts.rom_param, extra_waveform_kwargs=extra_waveform_args, **extra_args)
hlmT_1  = {}
for mode in hlmF_1:
    hlmT_1[mode] = lalsimutils.DataInverseFourier(hlmF_1[mode])
    print(mode, np.max(np.abs(hlmT_1[mode].data.data))*np.abs(lal.SpinWeightedSphericalHarmonic(P.incl,- P.phiref,-2,mode[0],mode[1])) ) #,"\t\t", 1./hlmF_1[mode].deltaF, hlmT_1[mode].deltaT*hlmT_1[mode].data.length)
P_copy  = P.manual_copy()
hTc_1 = lalsimutils.hoft_from_hlm(hlmT_1, P_copy,return_complex=True,extra_phase_shift=np.pi*opts.force_phase_shift_factor_of_pi)
if opts.verbose:
    print('net ', np.max(np.abs(hTc_1.data.data)))

# hoft direct. Currently lalsuite only
if not(opts.use_gwsignal):
  hTc_2  = lalsimutils.complex_hoft(P,extra_waveform_args=extra_waveform_args)
else:
  hTc_2 = rift_gws.complex_hoft(P, approx_string=opts.approximant, extra_waveform_args=extra_waveform_args)
  # import astropy.units as u
  # python_dict = {'mass1' : P.m1/lal.MSUN_SI * u.solMass,
  #             'mass2' : P.m2/lal.MSUN_SI * u.solMass,
  #             'spin1x' : P.s1x*u.dimensionless_unscaled,
  #             'spin1y' : P.s1y*u.dimensionless_unscaled,
  #             'spin1z' : P.s1z*u.dimensionless_unscaled,
  #             'spin2x' : P.s2x*u.dimensionless_unscaled,
  #             'spin2y' : P.s2y*u.dimensionless_unscaled,
  #             'spin2z' : P.s2z*u.dimensionless_unscaled,
  #             'deltaT' : P.deltaT*u.s,
  #             'f22_start' : P.fmin*u.Hz,
  #             'f22_ref': P.fref*u.Hz,
  #             'phi_ref' : P.phiref*u.rad,
  #             'distance' : P.dist/(1e6*lal.PC_SI)*u.Mpc,
  #             'inclination' : P.incl*u.rad,
  #             'eccentricity' : P.eccentricity*u.dimensionless_unscaled,
  #             'longAscNodes' : P.psi*u.rad,
  #             'meanPerAno' : P.meanPerAno*u.rad,
  #             'condition' : 1}
  # hp, hc = gws.core.waveform.GenerateTDWaveform(python_dict, gen)
  #hTc_2 = 
if opts.verbose:
  print('net2 ', np.max(np.abs(hTc_2.data.data)))

# now confirm complex_hoft dependence on psi is as desired
#    NOT THE SAME PSI DEPENDENCE AS WE ASSUME ELSEWHERE
psi_ref = float(P.psi)
P.psi = 0
if opts.use_gwsignal:
   hTc_3 = rift_gws.complex_hoft(P,approx_string=opts.approximant); P.psi = psi_ref
else:
   hTc_3 = lalsimutils.complex_hoft(P); P.psi = psi_ref
hTc_3.data.data *= np.exp(-2j*P.psi)

if opts.verbose:
    dh =np.max(np.abs(hTc_3.data.data - hTc_2.data.data))
    print(" Max diff psi - confirm psi coding correct ", dh )
    print(" Duration check ", len(hTc_1.data.data), len(hTc_2.data.data), len(hTc_3.data.data))

tvals1 = lalsimutils.evaluate_tvals(hTc_1)
tvals2 = lalsimutils.evaluate_tvals(hTc_2)
indx1 = np.argmax(np.abs(hTc_1.data.data))
indx2 = np.argmax(np.abs(hTc_2.data.data))

insert_string = str(P.approx)
if not(isinstance(P.approx, str)):
   import lalsimulation as lalsim
   insert_string = lalsim.GetStringFromApproximant(P.approx)


plt.title(opts.approximant)
if opts.verbose:
    print( tvals1[indx1], np.angle(hTc_1.data.data[indx1]), P.psi )
    print( tvals2[indx2], np.angle(hTc_2.data.data[indx2]), P.psi )
plt.plot(tvals1, np.abs(hTc_1.data.data),c='k')
plt.plot(tvals2, np.abs(hTc_2.data.data),c='r')
plt.savefig(f"wf_{opts.approximant}_long_check.png")
plt.plot(tvals1, np.real(hTc_1.data.data),c='k',lw=1)
plt.plot(tvals2, np.real(hTc_2.data.data),c='r',lw=1)
#if dh > 1e-4 *np.max
plt.plot(tvals2, np.real(hTc_3.data.data),c='g',lw=1)
#plt.xlim(np.min([tvals1[0],tvals2[0] ]), 0.1)
plt.xlim(opts.tmin,opts.tmax)
plt.title(opts.approximant)

plt.savefig(f"wf_{opts.approximant}_check.png")



from factored_likelihood import *
from matplotlib import pylab as plt

m1 = 10.*lal.LAL_MSUN_SI
m2 = 10.*lal.LAL_MSUN_SI
fmin_sig = 10.
fmin_tmplt = 40.
Nsamp = 500

# First determine the length of the waveform
# and choose P.tref so geocenter data starts at time 'gc_epoch'
gc_epoch = 100.0
Psig = ChooseWaveformParams(fmin=fmin_sig, radec=False, m1=m1, m2=m2)
df = findDeltaF(Psig)
Psig.deltaF = df
htest = hoft(Psig)
print("Length of the signal waveform is:", float(htest.epoch))
gc_tref = gc_epoch - float(htest.epoch)

IP = Overlap(fLow=40., deltaF=df, psd=lal.LIGOIPsd, full_output=True)

# Signal comes up through South pole, therefore, should arrive at
# geocenter, L1, V1, H1 in that order
RA = 0.
DEC = - lal.LAL_PI_2
Psig = ChooseWaveformParams(fmin=fmin_sig, radec=True, m1=m1, m2=m2,
        theta=DEC, phi=RA, detector='H1', tref=gc_tref, deltaF=df)

# Make a template of same masses, but shorter
Ptmplt = ChooseWaveformParams(fmin=fmin_tmplt, m1=m1, m2=m2, deltaF=df)

# Compute TD and normalized FD signals and template
htH1 = hoft(Psig)
hfH1 = norm_hoff(Psig, IP)
Psig.detector = 'L1' 
htL1 = hoft(Psig)
hfL1 = norm_hoff(Psig, IP)
Psig.detector = 'V1'
htV1 = hoft(Psig)
hfV1 = norm_hoff(Psig, IP)

htTMPLT = hoft(Ptmplt)
hfTMPLT = norm_hoff(Ptmplt, IP)

# Make array of discrete time steps for each detector
tH1 = float(htH1.epoch) + np.arange(htH1.data.length) * htH1.deltaT
tL1 = float(htL1.epoch) + np.arange(htL1.data.length) * htL1.deltaT
tV1 = float(htV1.epoch) + np.arange(htV1.data.length) * htV1.deltaT
tTMPLT = gc_epoch + np.arange(htTMPLT.data.length) * htTMPLT.deltaT

print("Template length is:", -float(htTMPLT.epoch))
print("H1 epoch is:", float(htH1.epoch))
print("\tso max should be at time:", float(htH1.epoch) - float(htest.epoch)\
        + float(htTMPLT.epoch))
print("L1 epoch is:", float(htL1.epoch))
print("\tso max should be at time:", float(htL1.epoch) - float(htest.epoch)\
        + float(htTMPLT.epoch))
print("V1 epoch is:", float(htV1.epoch))
print("\tso max should be at time:", float(htV1.epoch) - float(htest.epoch)\
        + float(htTMPLT.epoch))

rhoMaxH1, rhoH1, rhoIdxH1, rhoArgH1 = IP.ip(hfH1, hfTMPLT)
rhoMaxL1, rhoL1, rhoIdxL1, rhoArgL1 = IP.ip(hfL1, hfTMPLT)
rhoMaxV1, rhoV1, rhoIdxV1, rhoArgV1 = IP.ip(hfV1, hfTMPLT)

print("Max SNRs are:")
print("H1 max =", rhoMaxH1, "at index", rhoIdxH1, "or time", tH1[rhoIdxH1])
print("L1 max =", rhoMaxL1, "at index", rhoIdxL1, "or time", tL1[rhoIdxL1])
print("V1 max =", rhoMaxV1, "at index", rhoIdxV1, "or time", tV1[rhoIdxV1])

firstH1 = int(rhoIdxH1 - Nsamp)
lastH1 = rhoIdxH1 + Nsamp
firstL1 = int(rhoIdxL1 - Nsamp)
lastL1 = rhoIdxL1 + Nsamp
firstV1 = int(rhoIdxV1 - Nsamp)
lastV1 = rhoIdxV1 + Nsamp
rhoH1cut = lal.CutCOMPLEX16TimeSeries(rhoH1, firstH1, 2*Nsamp+1)
rhoL1cut = lal.CutCOMPLEX16TimeSeries(rhoL1, firstL1, 2*Nsamp+1)
rhoV1cut = lal.CutCOMPLEX16TimeSeries(rhoV1, firstV1, 2*Nsamp+1)

tH1cut = tH1[firstH1:lastH1+1]
tL1cut = tL1[firstL1:lastL1+1]
tV1cut = tV1[firstV1:lastV1+1]

print("Cut length", rhoH1cut.data.length)
print("Length of time steps", len(tH1cut))

# Plot TD template and H1 waveform
plt.figure(1)
plt.plot(tH1, htH1.data.data, 'b-')
plt.plot(tL1, htL1.data.data, 'g-')
plt.plot(tV1, htV1.data.data, 'm-')
plt.plot(tTMPLT, htTMPLT.data.data, 'k-')

# Plot discrete overlap vs time
plt.figure(2)
plt.plot(tH1, np.abs(rhoH1.data.data), 'b-')
plt.plot(tL1, np.abs(rhoL1.data.data), 'g-')
plt.plot(tL1cut, np.abs(rhoL1cut.data.data), 'k--')
plt.plot(tV1, np.abs(rhoV1.data.data), 'm-')

# interpolate the discrete overlaps
tintpH1 = np.arange(float(htH1.epoch),float(htH1.epoch)+1./df,10**-5)
tintpL1 = np.arange(float(htL1.epoch),float(htL1.epoch)+1./df,10**-5)
tintpV1 = np.arange(float(htV1.epoch),float(htV1.epoch)+1./df,10**-5)
ampH1 = np.abs(rhoH1.data.data)
phaseH1 = unwind_phase( np.angle(rhoH1.data.data) )
ampintpH1 = interpolate.InterpolatedUnivariateSpline(tH1, ampH1, k=3)
phaseintpH1 = interpolate.InterpolatedUnivariateSpline(tH1, phaseH1, k=3)
rhointpH1 = lambda x: ampintpH1(x) * np.exp( 1j * phaseintpH1(x) )
ampL1 = np.abs(rhoL1.data.data)
phaseL1 = unwind_phase( np.angle(rhoL1.data.data) )
ampintpL1 = interpolate.InterpolatedUnivariateSpline(tL1, ampL1, k=3)
phaseintpL1 = interpolate.InterpolatedUnivariateSpline(tL1, phaseL1, k=3)
rhointpL1 = lambda x: ampintpL1(x) * np.exp( 1j * phaseintpL1(x) )
ampV1 = np.abs(rhoV1.data.data)
phaseV1 = unwind_phase( np.angle(rhoV1.data.data) )
ampintpV1 = interpolate.InterpolatedUnivariateSpline(tV1, ampV1, k=3)
phaseintpV1 = interpolate.InterpolatedUnivariateSpline(tV1, phaseV1, k=3)
rhointpV1 = lambda x: ampintpV1(x) * np.exp( 1j * phaseintpV1(x) )

plt.figure(3)
plt.plot(tintpH1, np.abs( rhointpH1(tintpH1) ), 'b-')
plt.plot(tintpL1, np.abs( rhointpL1(tintpL1) ), 'g-')
plt.plot(tintpV1, np.abs( rhointpV1(tintpV1) ), 'm-')

# interpolate the truncated overlaps
tintpH1cut = np.arange(tH1cut[0],tH1cut[-1],10**-5)
tintpL1cut = np.arange(tL1cut[0],tL1cut[-1],10**-5)
tintpV1cut = np.arange(tV1cut[0],tV1cut[-1],10**-5)
ampH1cut = np.abs(rhoH1cut.data.data)
phaseH1cut = unwind_phase( np.angle(rhoH1cut.data.data) )
ampintpH1cut = interpolate.InterpolatedUnivariateSpline(tH1cut, ampH1cut, k=3)
phaseintpH1cut = interpolate.InterpolatedUnivariateSpline(tH1cut, phaseH1cut, k=3)
rhointpH1cut = lambda x: ampintpH1cut(x) * np.exp( 1j * phaseintpH1cut(x) )
ampL1cut = np.abs(rhoL1cut.data.data)
phaseL1cut = unwind_phase( np.angle(rhoL1cut.data.data) )
ampintpL1cut = interpolate.InterpolatedUnivariateSpline(tL1cut, ampL1cut, k=3)
phaseintpL1cut = interpolate.InterpolatedUnivariateSpline(tL1cut, phaseL1cut, k=3)
rhointpL1cut = lambda x: ampintpL1cut(x) * np.exp( 1j * phaseintpL1cut(x) )
ampV1cut = np.abs(rhoV1cut.data.data)
phaseV1cut = unwind_phase( np.angle(rhoV1cut.data.data) )
ampintpV1cut = interpolate.InterpolatedUnivariateSpline(tV1cut, ampV1cut, k=3)
phaseintpV1cut = interpolate.InterpolatedUnivariateSpline(tV1cut, phaseV1cut, k=3)
rhointpV1cut = lambda x: ampintpV1cut(x) * np.exp( 1j * phaseintpV1cut(x) )

plt.figure(4)
plt.plot(tintpH1cut, np.abs( rhointpH1cut(tintpH1cut) ), 'b-')
plt.plot(tintpL1cut, np.abs( rhointpL1cut(tintpL1cut) ), 'g-')
plt.plot(tintpV1cut, np.abs( rhointpV1cut(tintpV1cut) ), 'm-')
plt.show()





import sys
import math

import numpy
np=numpy #import numpy as np

import scipy.optimize

from RIFT.integrators import multivariate_truncnorm as truncnorm



from RIFT.integrators.mcsampler_generic import MCSamplerGeneric

class ClimbingOracle(MCSamplerGeneric):
    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.reference_samples = None
        self.lnL = None # function to climb
        self.n_climbers = None

    def add_parameter(self, params, pdf,   **kwargs):
        super().add_parameter(params,pdf, **kwargs)

    def setup(self, lnL=None,n_climbers = 10,**kwargs):
        super().setup(**kwargs)
        self.lnL= lnL
        self.n_climbers = n_climbers

        # create bounds formatting.  Need to follow syntax for scipy.optimize.minimize
        bounds = []
        for indx, p in enumerate(self.params_ordered):
            bounds.append(  (self.llim[p], self.rlim[p]) )

        self._bounds = bounds


    def update_sampling_prior(self,ln_weights, n_history,lnw_cut = -10, external_rvs=None,verbose=False,**kwargs):
        if verbose:
            print(" ---- ORACLE UPDATE - CLIMB --- ")
        # Allow updates provided from outside sources, assuming weights provided
        rvs_here = self._rvs
        if external_rvs:
            rvs_here = external_rvs

        # if no history
        if rvs_here is None:
            raise Exception(" oracle  - update  -update_sampling_prior requires initial history")

        n_history_to_use = np.min([n_history,  len(rvs_here[self.params_ordered[0]])] )
        if not(ln_weights is None):
            n_history_to_use = np.min([n_history,len(ln_weights)])
        
        sample_array = self.xpy.empty( (len(self.params_ordered), n_history_to_use))
        for indx, p in enumerate(self.params_ordered):
            sample_array[indx] = rvs_here[p][-n_history_to_use:]


        if lnw_cut and not(ln_weights is None): # we can override and trainon all data
            sample_array = sample_array[:, ln_weights > np.max(ln_weights) + lnw_cut ]  # training range

        # Pick points to climb
        drawn_indx = np.random.choice(range(len(sample_array)), replace=True,size=self.n_climbers) # random samples
        sample_array = sample_array[:,drawn_indx].T

        # Climb 
        rv_out = []
        def fn_here(x):
            xp = x.reshape(1,-1).T
            # lnL takes a *x as arguments, each an array. fn_here takes a single vector-valued argument of scalars
            try: 
                return - self.lnL(*xp)
            except:
                return -np.inf
        
        args_minimize = kwargs['minimize_args'] if 'minimize_args' in kwargs else {'tol':0.1, 'options': {'maxiter':5}}
        for indx in range(self.n_climbers):
            res = scipy.optimize.minimize(fn_here, sample_array[indx], **args_minimize)
            rv_out.append(res.x)

        # create samples based on draws
        self.reference_samples = np.array(rv_out)

    def draw_simplified(self, n_samples, *args, **kwargs):
        print(" ---- ORACLE DRAW --- ")
        rv_out = np.empty( (n_samples, len(self.params_ordered)))

        drawn_indx = np.random.choice(range(len(self.reference_samples)), size=n_samples) # random samples
        rv_out = self.reference_samples[drawn_indx]

        return None, None,rv_out  # compatible interface with mcsampler
        



#! /usr/bin/env python

import argparse
import numpy as np
from pesummary.gw.file.strain import StrainData
from gwpy.timeseries import TimeSeries
from gwpy.signal import filter_design
import matplotlib.pyplot as plt
import RIFT.lalsimutils as lalsimutils
import RIFT.physics.GWSignal as gws
import lal
import lalsimulation as lalsim
import os
import re
import sys
from pesummary.core.plots.interpolate import Bounded_interp1d
import matplotlib as mpl
mpl.rcParams["text.usetex"] = False

## USAGE
# python plot-waveform-ci.py --approx 'SEOBNRv5EHM' --samples 'extrinsic-1000-draws.dat' --event-name 'GW150914'

## TODO
# - if extrinsic file is default, take a downsample for speed??
# - can this be turned into a dag? waveform gen is slow

#################################################### 
### setup - user must supply event name, and change approx/samples file if desired
#################################################### 

parser = argparse.ArgumentParser()
parser.add_argument("--approx",default='SEOBNRv4',type=str,help="Approximant. REQUIRED")
parser.add_argument("--samples",default='extrinsic_posterior_samples.dat',type=str,help="Posterior samples used to generate 90% CI")
parser.add_argument("--downselect", action="store_true", help="Downselect number of waveforms to generate. Try 1000")
parser.add_argument("--draws",default=None,type=int,help="MUST be set if using downselect. Try 1000")
parser.add_argument("--event-name", default='GW',type=str,help="event name for plot titles and filenames")
opts=  parser.parse_args()

# defaults - can this be read in from anywhere??
event_name = opts.event_name

set_approx_str = opts.approx
samples = opts.samples

# assuming this is run in a standard RIFT directory
with open("ILE.sub", "r") as f:
    content = f.read()
match = re.search(r'--event-time\s+([\d.]+)', content)
if match:
    tevent = float(match.group(1))
    print(f"Extracted event time: {tevent}")
else:
    print("ERROR: event time not found.")
    sys.exit()

def get_random_lines(filename, num_lines):
    with open(filename) as f:
        lines = f.readlines()
    header = lines[0]
    data_lines = lines[1:]
    return header, random.sample(data_lines, num_lines)

## needs to be tested!!!
if opts.downselect:
    if not opts.draws > 0:
        print("ERROR: you must set draws is using downselect")
        sys.exit()
    header, random_lines = get_random_lines(samples, draws)
    with open(f'{samples}-downsample.dat',w) as f:
        f.write(header)
        for line in random_lines:
            f.write(line)
    samples = f'{samples}-downsample.dat'

# default waveform storage - avoid remaking this after initial build (slow)
waveform_filename = "waveform_realizations.npz"
    
####################################################
### Define necessary funcs - whiten, waveforms, CI
####################################################

# as in pesummary, bandpass/filter the data
# see gwosc tutorial
def whiten_strain(data, bp_freqs=[50,250], srate=4096.0, notches=[60., 120.,180.]):
    bp = filter_design.bandpass(*bp_freqs, srate)
    zpks = [filter_design.notch(line, srate) for line in notches]
    zpk = filter_design.concatenate_zpks(bp, *zpks)
    hfilt = data.filter(zpk, filtfilt=True)
    _strain = hfilt.crop(*hfilt.span.contract(1))
    return _strain

# make param list from sample file AND generate tvals/hoft PER detector - store them
def generate_waveform_realizations(samples_file, tevent, fmin=10., fref=10., approx=set_approx_str):
    samples = np.genfromtxt(samples_file, names=True, replace_space=None)
    P_list = []
    # so manual :( but it works
    for indx in np.arange(len(samples["m1"])):
        P = lalsimutils.ChooseWaveformParams()
        P.m1 = samples["m1"][indx] * lal.MSUN_SI
        P.m2 = samples["m2"][indx] * lal.MSUN_SI
        P.s1x = samples["a1x"][indx]
        P.s1y = samples["a1y"][indx]
        P.s1z = samples["a1z"][indx]
        P.s2x = samples["a2x"][indx]
        P.s2y = samples["a2y"][indx]
        P.s2z = samples["a2z"][indx]

        # ecc/mpa not always present
        if 'eccentricity' in samples.dtype.names:
            P.eccentricity = samples["eccentricity"][indx]
        if 'meanPerAno' in samples.dtype.names:
            P.meanPerAno = samples["meanPerAno"][indx]

        P.ampO = -1
        P.dist = samples["distance"][indx] * lal.PC_SI * 1.e6
        P.incl = samples["incl"][indx]
        P.theta = samples["dec"][indx]
        P.phi = samples["ra"][indx]
        P.psi = samples["psi"][indx]
        P.phiref = samples["phiorb"][indx] - np.pi/2
        P.tref = samples["time"][indx]
        P.radec = True
        P.detector = "H1"
        P.approx = approx
        P.deltaF = 1./16
        P.fmin = fmin
        P.fref = fref
        P_list.append(P)

    # generate waveforms for each set of parameters
    # double check, some redundancy here
    tvals_H_list, ht_H_list, tvals_L_list, ht_L_list = [], [], [], []
    for param in P_list:

        if param.approx.startswith('SEOBNR'):
            hoft_H = gws.hoft(param, approx_string=param.approx)
            param.detector = "L1"
            hoft_L = gws.hoft(param, approx_string=param.approx)
        else:
            param.approx = lalsim.GetApproximantFromString(param.approx)
            hoft_H = lalsimutils.hoft(param)
            param.detector = "L1"
            hoft_L = lalsimutils.hoft(param)
            
        tvals_H = lalsimutils.evaluate_tvals(hoft_H) - tevent
        indx_ok = np.logical_and(tvals_H > -0.5, tvals_H < 0.2)
        tvals_H = tvals_H[indx_ok]
        hoft_H = TimeSeries(hoft_H.data.data[indx_ok], name='Strain', t0=tvals_H[0], unit='dimensionless', dt=tvals_H[1] - tvals_H[0])
        hoft_H = whiten_strain(hoft_H)
        tvals_H_list.append(tvals_H)
        ht_H_list.append(hoft_H)

        tvals_L = lalsimutils.evaluate_tvals(hoft_L) - tevent
        indx_ok = np.logical_and(tvals_L > -0.5, tvals_L < 0.2)
        tvals_L = tvals_L[indx_ok]
        hoft_L = TimeSeries(hoft_L.data.data[indx_ok], name='Strain', t0=tvals_L[0], unit='dimensionless', dt=tvals_L[1] - tvals_L[0])
        hoft_L = whiten_strain(hoft_L)
        tvals_L_list.append(tvals_L)
        ht_L_list.append(hoft_L)

    return tvals_H_list, ht_H_list, tvals_L_list, ht_L_list

# generic function to compute the 90% CIs
# done as pesummary/gwosc tutorial
def compute_ci(tvals, waveform_realizations):
    new_t = np.arange(min(tvals[0]), max(tvals[0]), 1/4096.)

    td_waveform_array = np.array([
        Bounded_interp1d(times, amplitude, xlow=min(tvals[0]), xhigh=max(tvals[0]))(new_t)
        for times, amplitude in zip(tvals, waveform_realizations)
    ])
    
    upper = np.percentile(td_waveform_array, 95, axis=0)
    lower = np.percentile(td_waveform_array, 5, axis=0)

    return new_t, upper, lower

####################################################
### get data and call functions to get waveforms ###
####################################################

# fetch open data, pesummary
H1_data = StrainData.fetch_open_data('H1', tevent - 20, tevent + 5)
L1_data = StrainData.fetch_open_data('L1', tevent - 20, tevent + 5)
# whiten
_strain_H = whiten_strain(H1_data)
times_H = [t-tevent for t in np.array(_strain_H.times)]
_strain_L = whiten_strain(L1_data)
times_L = [t-tevent for t in np.array(_strain_L.times)]

# do not rebuild waveforms for CI unless you have to
if os.path.exists(waveform_filename):
    print(f"File {waveform_filename} found. Loading existing waveforms.")

else:
    print(f"File {waveform_filename} not found. Generating waveforms.")
    tvals_H_list, ht_H_list, tvals_L_list, ht_L_list = generate_waveform_realizations(samples, tevent)
    np.savez(waveform_filename, 
             times_H=np.array(tvals_H_list), amplitude_H=np.array(ht_H_list),
             times_L=np.array(tvals_L_list), amplitude_L=np.array(ht_L_list))

# for consistency, load and use file either way
data = np.load(waveform_filename)
tvals_H = data["times_H"]
waveform_realizations_H = data["amplitude_H"]
tvals_L = data["times_L"]
waveform_realizations_L = data["amplitude_L"]

# get 90% CIs
new_t_H, upper_H, lower_H = compute_ci(tvals_H, waveform_realizations_H)
new_t_L, upper_L, lower_L = compute_ci(tvals_L, waveform_realizations_L)

####################################################
###  Make the plots ###
####################################################

# plot strain data
plt.figure(figsize=(20,5))
plt.plot(times_H,_strain_H,'gray',label='Real H1 strain',alpha=0.7,lw=2.5)
plt.plot(times_L,_strain_L,'gray',label='Real L1 strain',alpha=0.4,lw=2.5)
plt.xlim([-0.3,0.075])
plt.xlabel('Time (seconds) from '+str(tevent))
plt.ylabel('Strain')
plt.legend(loc='lower left')
plt.title(f'LIGO strain data near {event_name}')
plt.savefig(f"{event_name}_strain_data.png")

# plot h(t) waveforms, H as an example
plt.figure(figsize=(15,5))
for i in range(0,len(waveform_realizations_H)):
    plt.plot(np.array(tvals_H[i]), waveform_realizations_H[i],'gray',alpha=0.25)
plt.xlabel('Time (seconds) from '+str(tevent))
plt.ylabel('Strain')
plt.title('Waveform Comparison');
plt.xlim(-0.3,0.075);
plt.savefig(f"{event_name}_ht_waveforms.png", bbox_inches='tight')

## ci + data h1/l1 sep panels
fig, axes = plt.subplots(2, 1, figsize=(15, 10), sharex=True)

axes[0].plot(times_H, _strain_H, 'gray', label='Real H1 strain', alpha=0.4, lw=2.5)
axes[0].fill_between(new_t_H, upper_H, lower_H, color='lightgreen', alpha=0.4, label=f'{set_approx_str} 90% CI')
axes[0].set_xlim([-0.3, 0.075])
axes[0].set_ylabel('strain')
axes[0].legend(loc='lower left')
axes[0].set_title(f'LIGO strain data near {event_name}',fontsize=24)
axes[0].text(.5,.9,'LIGO Hanford',horizontalalignment='center',transform=axes[0].transAxes)

axes[1].plot(times_L, _strain_L, 'gray', label='Real L1 strain', alpha=0.4, lw=2.5)
axes[1].fill_between(new_t_L, upper_L, lower_L, color='lightgreen', alpha=0.4, label=f'{set_approx_str} 90% CI')
axes[1].set_xlim([-0.3, 0.075])
axes[1].set_xlabel('time (seconds) from ' + str(tevent))
axes[1].set_ylabel('strain')
axes[1].legend(loc='lower left')
axes[1].text(.5,.9,'LIGO Livingston',horizontalalignment='center',transform=axes[1].transAxes)

plt.tight_layout()
plt.savefig(f"{event_name}_CI_H1L1.png", bbox_inches='tight')



#! /bin/env python
#


import spokes
import numpy as np
from matplotlib import pyplot as plt


#xvals = np.linspace(-1,-0.5,30)
xvals = np.linspace(-1,0.1,30)
fnHere = lambda x: 200*np.cos(x)
yvals =fnHere(xvals)   # Note lnL scale is an option, need to be careful to choose a reasonable default

plt.plot(xvals,yvals,'o')

code, xvals_new = spokes.Refine(xvals,yvals,npts=30)
print(code,len(xvals_new))
if code == 'refined' or code=='extended':
    yvals_new = fnHere(xvals_new)
    plt.plot(xvals_new,yvals_new,'r')

plt.show()



#! /usr/bin/env python

# Author: Askold Vilkha (https://github.com/askoldvilkha), av7683@rit.edu, askoldvilkha@gmail.com

# This script is used to plot the results of the tabular EOS inference. 
# Minimal input is the path to the tabular EOS file and posterior samples files the user wants to have on the plot.
# The user will have to specify which plots to make by setting the corresponding flags to True.
# The script has a functionality to choose custom labels and colors for the posterior samples.
# If multiple tabular EOS files are provided, the user will have to specify which ones to plot and to use for the posterior samples.

import numpy as np
import matplotlib.pyplot as plt
import h5py
import RIFT.physics.EOSManager as EOSManager
#from natsort import natsorted
import warnings
import argparse

import ast
import textwrap

import RIFT.plot_utilities.TabularEOSPlotUtilities as tabplot

# store long help messages in a dictionary to avoid cluttering the parser code below
argparse_help_dict = {
    'help': textwrap.dedent('''\
    This script is used to plot the results of the tabular EOS inference obtained with RIFT code.
    The user can plot pressure vs. density and mass vs. radius plots for the tabular EOS priors and posterior samples.
    The user can specify the tabular EOS files, posterior samples files, labels, and colors for the plots.
    The user has to provide the path to the tabular EOS files and the posterior samples files.
    Basic usage: 
    plot_tabular_eos_inference.py --tabular-eos-file <path_to_tabular_eos_file> --tabular-eos-label <tabular_eos_file_label> --posterior-file <path_to_posterior_samples_file> 
    --posterior-label <posterior_samples_file_label> --plot-p-vs-rho --plot-m-vs-r --use-bgcgb-colormap'''),

    'tabular-eos-file': textwrap.dedent('''\
    Path to the tabular EOS file. If multiple files are provided only the first one will be used for EOS posteriors unless specified otherwise by the user. 
    See --posterior-tabular-map for more information.','''),

    'plot-tabular-eos-prior': textwrap.dedent('''\
    Specify which tabular EOS files to plot priors for. 
    Should be specified by the label of the tabular EOS file'''),

    'posterior-tabular-map': textwrap.dedent('''\
    Specify which tabular EOS file to use for the posterior samples. 
    Should be specified by the label of the tabular EOS files and the posterior samples files. 
    The input format should be a dictionary in the form of a string:
    '{'tabular_eos_file_1': ['posterior_file_1', 'posterior_file_2'] 'tabular_eos_file_2': 'posterior_file_3', ...}'. 
    Use either labels instead of the file names.'''),

    'posterior-histogram-range-r': textwrap.dedent('''\
    Set the range for the Radius at the posterior samples histogram plot. 
    The input format should be a list of two numbers defining a range for the Radius. (e.g. [10, 15])'''),

    'posterior-histogram-range-lambda': textwrap.dedent('''\
    Set the range for the Tidal Deformability Lambda at the posterior samples histogram plot.
    The input format should be a list of two numbers defining a range for the Tidal Deformability Lambda. (e.g. [400, 1600])''')
}

parser = argparse.ArgumentParser(description = argparse_help_dict['help'])

# basic arguments
parser.add_argument('--tabular-eos-file', action = 'append', help = argparse_help_dict['tabular-eos-file'], required = True)
parser.add_argument('--posterior-file', action = 'append', help = 'Path to the posterior samples file. If none are provdided, only priors will be plotted.')
parser.add_argument('--plot-p-vs-rho', action = 'store_true', help = 'Plot pressure vs. density')
parser.add_argument('--plot-m-vs-r', action = 'store_true', help = 'Plot mass vs. radius')
parser.add_argument('--tabular-eos-label', action = 'append', help = 'Label for the tabular EOS file')
parser.add_argument('--posterior-label', action = 'append', help = 'Label for the posterior samples file')
parser.add_argument('--color', action = 'append', help = 'Colors for the plot. If not provided, colors will be chosen automatically')
parser.add_argument('--use-bgcgb-colormap', action = 'store_true', help = 'Use the BlackGreyCyanGreenBlue colormap for the plots')
parser.add_argument('--verbose', action = 'store_true', help = 'Print information on the progress of the code')
parser.add_argument('--plot-p-vs-rho-title', action = 'store', help = 'Title for the pressure vs. density plot')
parser.add_argument('--plot-m-vs-r-title', action = 'store', help = 'Title for the mass vs. radius plot')

# extra arguments for more advanced plotting (plot posteriors for multiple tabular EOS files, plot tabular EOS priors, etc.)
parser.add_argument('--plot-tabular-eos-prior', action = 'append', help = argparse_help_dict['plot-tabular-eos-prior'])
parser.add_argument('--posterior-tabular-map', action = 'store', help = argparse_help_dict['posterior-tabular-map'])
parser.add_argument('--plot-posterior-histogram', action = 'store_true', help = 'Plot the posterior samples histogram')
parser.add_argument('--plot-lambda-tilde-ratio', action = 'store_true', help = 'Plot the lambda_tilde vs ordering statistics S ratio')
parser.add_argument('--posterior-histogram-range-r', action = 'store', help = 'Range for Radius at the posterior samples histogram plot')
parser.add_argument('--posterior-histogram-range-lambda', action = 'store', help = 'Range for Tidal Deformability Lambda at the posterior samples histogram plot')

args = parser.parse_args()

tabular_eos_labels = [None] * len(args.tabular_eos_file)

# the code can operate with less labels than files, it will use the default labels for the rest of the files
if args.tabular_eos_label is not None:
    if len(args.tabular_eos_label) < len(args.tabular_eos_file):
        warnings.warn('Number of tabular EOS labels is less than the number of tabular EOS files. The code will use the default labels after the last provided label.')
    elif len(args.tabular_eos_label) > len(args.tabular_eos_file):
        raise ValueError('Number of tabular EOS labels is greater than the number of tabular EOS files. Please check if you have provided the correct number of labels or files.')

    tabular_eos_labels[:len(args.tabular_eos_label)] = args.tabular_eos_label # fill in the provided labels, use the default labels for the rest

    if args.verbose:
        print('\nChecked the tabular EOS file labels, proceeding with the data loading...')

# posterior histogram and lambda_tilde ratio plots require saving the EOS manager object and/or the posterior samples (not just eos_indx column)
save_eos_manager = False
save_posterior_samples = False
if args.plot_posterior_histogram or args.plot_lambda_tilde_ratio:
    if args.posterior_file is None:
        raise ValueError('You have chosen to plot the posterior samples histogram or lambda_tilde ratio without providing the posterior samples files. Please provide the posterior samples files.')
    save_eos_manager = True
if args.plot_lambda_tilde_ratio:
    save_posterior_samples = True

# load the tabular EOS files
tabular_eos_data = {}
for i, tabular_eos_file in enumerate(args.tabular_eos_file):
    eos_data_i = tabplot.EOS_data_loader(tabular_eos_file, tabular_eos_labels[i], args.verbose, save_eos_manager)
    eos_data_i_label = eos_data_i['data_label'] # if the label was not provided, the EOS_data_loader function will assign a default label
    tabular_eos_data[eos_data_i_label] = eos_data_i
    tabular_eos_labels[i] = eos_data_i_label # update the label in case the default label was assigned

if args.verbose:
    print('\nLoaded the tabular EOS data.')
    for i in range(len(tabular_eos_labels)):
        print(f'Tabular EOS file: {args.tabular_eos_file[i]} with label: {tabular_eos_labels[i]}')

if args.posterior_file is not None:
    priors_labels = []
    posterior_samples_labels = [None] * len(args.posterior_file)

    if args.posterior_label is not None:
        if len(args.posterior_label) < len(args.posterior_file):
            warnings.warn('Number of posterior samples labels is less than the number of posterior samples files. The code will use the default labels after the last provided label.')
        elif len(args.posterior_label) > len(args.posterior_file):
            raise ValueError('Number of posterior samples labels is greater than the number of posterior samples files. Please check if you have provided the correct number of labels or files.')

        posterior_samples_labels[:len(args.posterior_label)] = args.posterior_label 
    
        if args.verbose:
            print('\nChecked the posterior samples labels, proceeding with the data loading...')

    # load the posterior samples files
    posterior_samples_data = {}
    for i, posterior_file in enumerate(args.posterior_file):
        posterior_data_i = tabplot.posterior_data_loader(posterior_file, posterior_samples_labels[i], args.verbose, save_posterior_samples)
        posterior_data_i_label = posterior_data_i['data_label']
        posterior_samples_data[posterior_data_i_label] = posterior_data_i
        posterior_samples_labels[i] = posterior_data_i_label

    if args.verbose:
        print('\nLoaded the posterior samples data.')
        for i in range(len(posterior_samples_labels)):
            print(f'Posterior samples file: {args.posterior_file[i]} with label: {posterior_samples_labels[i]}')

    # make the posterior-tabular map if not provided by the user
    if args.posterior_tabular_map is None:
        # if the user did not provide the posterior-tabular map, the code will only use first tabular EOS file for the posterior samples
        posterior_tabular_map = {tabular_eos_labels[0]: posterior_samples_labels}

    # check if the map is provided in the correct format and all the labels are valid
    else:
        if args.verbose:
            print('\nChecking the posterior-tabular map...')
        posterior_tabular_map = ast.literal_eval(args.posterior_tabular_map)
        if not isinstance(posterior_tabular_map, dict):
            raise ValueError('The posterior-tabular map should be a dictionary. Please check the format of the input.')
        for tabular_label in posterior_tabular_map.keys():
            if tabular_label not in tabular_eos_labels:
                raise ValueError(f'Tabular EOS label: {tabular_label} from your posterior-tabular map is not found in the tabular EOS labels. Please check the labels you provided.')
            if isinstance(posterior_tabular_map[tabular_label], list):
                for posterior_label in posterior_tabular_map[tabular_label]:
                    if posterior_label not in posterior_samples_labels:
                        raise ValueError(f'Posterior samples label: {posterior_label} from your posterior-tabular map is not found in the posterior samples labels. Please check the labels you provided.')
            else:
                if posterior_tabular_map[tabular_label] not in posterior_samples_labels:
                    raise ValueError(f'Posterior samples label: {posterior_tabular_map[tabular_label]} from your posterior-tabular map is not found in the posterior samples labels. Please check the labels you provided.')

    if args.verbose:
        print('\nCreated the posterior-tabular map.')
        for tabular_label in posterior_tabular_map.keys():
            print(f'Tabular EOS label: {tabular_label} with posterior samples labels: {posterior_tabular_map[tabular_label]}')
        print('\nMoving on to priors...')
else:
    posterior_samples_data = None
    posterior_tabular_map = None
    priors_labels = tabular_eos_labels # plot priors for all tabular EOS files if no posterior samples are provided
    warnings.warn('No posterior samples files provided. The code will only plot the tabular EOS priors.')
    
# if the user specified the tabular EOS files to plot priors for, the code will plot priors for those files
if args.plot_tabular_eos_prior is not None:
    for tabular_label in args.plot_tabular_eos_prior:
        if tabular_label not in tabular_eos_labels:
            raise ValueError(f'Tabular EOS label for prior plot: {tabular_label} is not found in the tabular EOS labels. Please check the labels you provided.')
    priors_labels = args.plot_tabular_eos_prior
elif len(tabular_eos_labels) > 1 and posterior_samples_data is not None:
    priors_labels = tabular_eos_labels[1:] # plot priors for all tabular EOS files except the first one

if args.verbose:
    print('\nChecked the tabular EOS labels for the priors.')
    for i in range(len(priors_labels)):
        print(f'Priors will be plotted for the tabular EOS file with label: {priors_labels[i]}')
    print('\nLinking the tabular EOS data with the posterior samples data according to the posterior-tabular map...')

plt.rcParams['figure.figsize'] = [15, 10]
plt.rcParams['font.size'] = 12

def EOS_plotter(args, tabular_eos_data: dict, posterior_samples_data: dict, posterior_tabular_map: dict, priors_labels: list, plot_type: str) -> None:
    """
    Function to plot the EOS inference results for the tabular EOS priors and posterior samples.
    
    Parameters:
    ----------
    args : argparse.Namespace
        The argparse namespace object with the script arguments.
    tabular_eos_data : dict
        The dictionary with the tabular EOS data generated earlier in the script. 
        Format: {tabular_label: tabular_eos_data}, tabular_eos_data should be generated by the EOS_data_loader function.
    posterior_samples_data : dict
        The dictionary with the posterior samples data generated earlier in the script. 
        Format: {posterior_label: posterior_samples_data}, posterior_samples_data should be generated by the posterior_data_loader function.
    posterior_tabular_map : dict
        The dictionary with the posterior-tabular map generated earlier in the script. 
        Format: {tabular_label: posterior_samples_label}, posterior_samples_label can be a list of labels or a single string label.
    priors_labels : list
        The list with the tabular EOS labels to plot priors for.
    plot_type : str
        The type of the plot to make. Can be 'pressure_density' or 'mass_radius'.

    Returns:
    -------
    None

    Raises:
    -------
    NOTE!
    Errors will not be raised if the script has not been changed from the initial version. 
    They are here to ensure future modifications do not break the code.

    ValueError
        If the plot type is not 'pressure_density' or 'mass_radius'. 
    ValueError
        If the posterior samples data is provided without the posterior-tabular map or vice versa.
    """
    if plot_type not in ['pressure_density', 'mass_radius']:
        raise ValueError('The plot type should be either pressure_density or mass_radius. Please check the input.')
    if (posterior_samples_data is None) != (posterior_tabular_map is None):
        raise ValueError('The posterior samples data and the posterior-tabular map should be provided together. Please check the input.')

    raw_plot_data = []

    for tabular_label in priors_labels:
        raw_plot_data.append(tabplot.link_eos_data_to_posterior(tabular_eos_data[tabular_label], plot_type))
    
    if args.verbose:
        print(f'\nCollected the data for priors for the {plot_type} plot. Moving on to posteriors...')
    
    # link the tabular EOS data with the posterior samples data according to the posterior-tabular map
    if posterior_samples_data is not None:
        for tabular_label in posterior_tabular_map.keys():
            if isinstance(posterior_tabular_map[tabular_label], list):
                for posterior_label in posterior_tabular_map[tabular_label]:
                    raw_plot_data.append(tabplot.link_eos_data_to_posterior(tabular_eos_data[tabular_label], plot_type, posterior_samples_data[posterior_label]))
            else:
                raw_plot_data.append(tabplot.link_eos_data_to_posterior(tabular_eos_data[tabular_label], plot_type, posterior_samples_data[posterior_tabular_map[tabular_label]]))

        if args.verbose:
            print(f'\nCollected the data for posterior samples for the {plot_type} plot. Processing...')
    
    # choose the colors for the plots
    colors = [None] * len(raw_plot_data)
    if args.color is not None:
        if len(args.color) < len(raw_plot_data):
            warnings.warn('Number of colors is less than the number of posterior samples files. The code will use the default colors for the rest of the files.')
            colors[:len(args.color)] = args.color
        else: 
            colors = args.color 
    if args.use_bgcgb_colormap:
        colors[:5] = ['black', 'grey', 'cyan', 'green', 'blue']
        if args.color is not None:
            warnings.warn('You have chosen to use the BlackGreyCyanGreenBlue colormap. The code will ignore the first 5 colors you provided.')
    if args.verbose and any(colors is not None for c in colors):
        print(f'\nThe colors for the plots are: {colors}. Any of the None values will be replaced with the default colors. (tab10 colormap)')
    
    plot_data = []
    if plot_type == 'pressure_density':
        for i, raw_data in enumerate(raw_plot_data):
            plot_data.append(tabplot.pressure_density_plot_data_gen(**raw_data, custom_color = colors[i]))
        
        if args.verbose:
            print(f'\nProcessed the data for the {plot_type} plot. Plotting...')

        tabplot.pressure_density_plot(*plot_data, title = args.plot_p_vs_rho_title)
    
    elif plot_type == 'mass_radius':
        for i, raw_data in enumerate(raw_plot_data):
            plot_data.append(tabplot.mass_radius_plot_data_gen(**raw_data, custom_color = colors[i]))

        if args.verbose:
            print(f'\nProcessed the data for the {plot_type} plot. Plotting...')

        tabplot.mass_radius_plot(*plot_data, title = args.plot_m_vs_r_title)

    return None
    
# plot the P vs rho EOS inference
if args.plot_p_vs_rho:
    EOS_plotter(args, tabular_eos_data, posterior_samples_data, posterior_tabular_map, priors_labels, 'pressure_density')

# plot the M vs R EOS inference
if args.plot_m_vs_r:
    EOS_plotter(args, tabular_eos_data, posterior_samples_data, posterior_tabular_map, priors_labels, 'mass_radius')

# plot the posterior samples histogram
if args.plot_posterior_histogram:
    
    posterior_hist_data = []
    for tabular_label in posterior_tabular_map.keys():
        if isinstance(posterior_tabular_map[tabular_label], list):
            for posterior_label in posterior_tabular_map[tabular_label]:
                posteriors_eos = posterior_samples_data[posterior_label]['posterior_samples_eos']
                eos_manager = tabular_eos_data[tabular_label]['eos_manager']
                posterior_hist_data.append(tabplot.posterior_hist_data_gen(posteriors_eos, eos_manager, posterior_label))
        else:
            posteriors_eos = posterior_samples_data[posterior_tabular_map[tabular_label]]['posterior_samples_eos']
            eos_manager = tabular_eos_data[tabular_label]['eos_manager']
            posterior_label = posterior_tabular_map[tabular_label]
            posterior_hist_data.append(tabplot.posterior_hist_data_gen(posteriors_eos, eos_manager, posterior_label))
    
    if args.verbose:
        print('\nProcessed the data for the posterior samples histogram. Plotting...')


    r_lim = (8, 17)
    lambda_lim = (0, 1000)
    if args.posterior_histogram_range_r is not None:
        r_lim_input = ast.literal_eval(args.posterior_histogram_range_r)
        r_lim = (r_lim_input[0], r_lim_input[1])
    if args.posterior_histogram_range_lambda is not None:
        lambda_lim_input = ast.literal_eval(args.posterior_histogram_range_lambda)
        lambda_lim = (lambda_lim_input[0], lambda_lim_input[1])

    tabplot.posterior_hist_plot(*posterior_hist_data, r_lim = r_lim, lambda_lim = lambda_lim)

# plot the lambda_tilde vs S ratio
if args.plot_lambda_tilde_ratio:
    
    lambda_tilde_ratio_data = []
    for tabular_label in posterior_tabular_map.keys():
        if isinstance(posterior_tabular_map[tabular_label], list):
            for posterior_label in posterior_tabular_map[tabular_label]:
                posteriors_samples_all = posterior_samples_data[posterior_label]['posterior_samples_all']
                eos_manager = tabular_eos_data[tabular_label]['eos_manager']
                lambda_tilde_ratio_data.append(tabplot.LambdaTilderatio_data_gen(posteriors_samples_all, eos_manager, posterior_label))
        else:
            posteriors_samples_all = posterior_samples_data[posterior_tabular_map[tabular_label]]['posterior_samples_all']
            eos_manager = tabular_eos_data[tabular_label]['eos_manager']
            posterior_label = posterior_tabular_map[tabular_label]
            lambda_tilde_ratio_data.append(tabplot.LambdaTilderatio_data_gen(posteriors_samples_all, eos_manager, posterior_label))
    
    if args.verbose:
        print('\nProcessed the data for the lambda_tilde vs S ratio plot. Plotting...')
    
    tabplot.LambdaTilderatio_plot(*lambda_tilde_ratio_data)
    tabplot.LambdaTilderatio_plot(*lambda_tilde_ratio_data, ylim = (0.8, 1.2))



#! /usr/bin/env python
#
# util_NRExtrudeOverlapGrid.py
#
#   This takes the whole NR catalog (aligned spin only!) and extrudes it in mtot (only)
#   Comparisons for cutoff are done ONLY using the 'approx' in LAL, for convenience
#
# EXAMPLES
#
#     # grid targeted to a single simulation
#        python util_NRExtrudeOverlapGrid.py --inj mdc.xml.gz  --group Sequence-SXS-All --param 1 --event 0 --verbose --skip-overlap
#     # extend the grid, to include the rest of the simlations
#        python util_NRExtrudeOverlapGrid.py --inj overlap-grid.xml.gz  --group Sequence-SXS-All  --event 0 --verbose --skip-overlap --insert-missing-spokes
#
# WARNINGS: i
#   xi_factor: meaning of grid is CHANGED from original interpretation
#


import argparse
import sys
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal
import functools

import RIFT.physics.effectiveFisher  as eff   # for the mesh grid generation
import RIFT.physics.PrecessingFisherMatrix   as pcf   # Superior tools to perform overlaps. Will need to standardize with Evans' approach in effectiveFisher.py

from multiprocessing import Pool
try:
    import os
    n_threads = int(os.environ['OMP_NUM_THREADS'])
    print(" Pool size : ", n_threads)
except:
    n_threads=1
    print(" - No multiprocessing - ")

try:
    import NRWaveformCatalogManager3 as nrwf
    hasNR =True
except:
    hasNR=False

try:
    hasEOB=True
    import EOBTidalExternal as eobwf
except:
    hasEOB=False


###
### Load options
###

parser = argparse.ArgumentParser()
# Parameters
parser.add_argument("--mtot-range",default='[50,110]')
parser.add_argument("--mc-range",default=None, help='If specified, overrides the total mass range [24,38]')
parser.add_argument("--grid-cartesian-npts",default=30)
parser.add_argument("--group",default=None)
parser.add_argument("--param", action='append', help='Explicit list of parameters to use')
parser.add_argument("--insert-missing-spokes", action='store_true')
parser.add_argument("--aligned-only",action='store_true')
parser.add_argument("--nospin-only",action='store_true')
parser.add_argument("--eccentricity",action='store_true')
parser.add_argument("--eta-range",default='[0.1,0.25]')
parser.add_argument("--mass-xi-factor",default=0.0,type=float, help="The mass ranges are assumed to apply at ZERO SPIN. For other values of xi, the mass ranges map to m_{used} = m(1+xi *xi_factor+(1/4-eta)*eta_factor).  Note that to be stable, xi_factor<1. Default value 0.6, based on relevant mass region")
parser.add_argument("--mass-eta-factor",default=0,type=float, help="The mass ranges are assumed to apply at ZERO SPIN. For other values of xi, the mass ranges map to m_{used} = m(1+xi *xi_factor+(1/4-eta)*eta_factor).  Note that to be stable, xi_factor<1. Default value 0.6, based on relevant mass region")
# Cutoff options
parser.add_argument("--skip-overlap",action='store_true', help="If true, the grid is generated without actually performing overlaps. Very helpful if the grid is just in mtot, for the purposes of reproducing a specific NR simulation")
parser.add_argument("--match-value", type=float, default=0.01, help="Use this as the minimum match value. Default is 0.01 (i.e., keep almost everything)")
# Overlap options
parser.add_argument("--fisher-psd",type=str,default="SimNoisePSDaLIGOZeroDetHighPower",help="psd name ('eval'). lalsim., lalsim.SimNoisePSDaLIGOZeroDetHighPower, lalsimutils.Wrapper_AdvLIGOPsd, .SimNoisePSDiLIGOSRD... ")
parser.add_argument("--psd-file",  help="File name for PSD (assumed hanford). Overrides --fisher-psd if provided")
parser.add_argument("--srate",type=int,default=16384,help="Sampling rate")
parser.add_argument("--seglen", type=float,default=64., help="Default window size for processing. Short for NR waveforms")
parser.add_argument("--fref",type=float,default=0.);
# External grid
parser.add_argument("--external-grid-xml", default=None,help="Inspiral XML file (injection form) for alternate grid")
parser.add_argument("--external-grid-txt", default=None, help="Cartesian grid. Must provide parameter names in header. Exactly like output of code. Last column not used.")
# Base point
parser.add_argument("--inj", dest='inj', default=None,help="inspiral XML file containing the base point (OR containing the reference XML grid to check against a database)")
parser.add_argument("--event",type=int, dest="event_id", default=None,help="event ID of injection XML to use.")
parser.add_argument("--fmin", default=10,type=float,help="Mininmum frequency in Hz, default is 40Hz to make short enough waveforms. Focus will be iLIGO to keep comutations short")
parser.add_argument("--require-fmin-above-NR-start",action='store_true')
parser.add_argument("--fmax",default=2000,type=float,help="Maximum frequency in Hz, used for PSD integral.")
parser.add_argument("--mass1", default=35,type=float,help="Mass in solar masses")  # 150 turns out to be ok for Healy et al sims
parser.add_argument("--mass2", default=35,type=float,help="Mass in solar masses")
parser.add_argument("--s1z", default=0.1,type=float,help="Spin1z")
#parser.add_argument("--lambda1",default=590,type=float)
#parser.add_argument("--lambda2", default=590,type=float)
parser.add_argument("--eff-lambda", type=float, help="Value of effective tidal parameter. Optional, ignored if not given")
parser.add_argument("--deff-lambda", type=float, help="Value of second effective tidal parameter. Optional, ignored if not given")
parser.add_argument("--lmax", default=2, type=int)
parser.add_argument("--approx",type=str,default="SEOBNRv2")
# Output options
parser.add_argument("--fname", default="overlap-grid", help="Base output file for ascii text (.dat) and xml (.xml.gz)")
parser.add_argument("--verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
parser.add_argument("--save-plots",default=False,action='store_true', help="Write plots to file (only useful for OSX, where interactive is default")
opts=  parser.parse_args()

xi_factor = opts.mass_xi_factor
eta_factor0 = opts.mass_eta_factor

if opts.verbose:
    True
    #lalsimutils.rosDebugMessagesContainer[0]=True   # enable error logging inside lalsimutils



###
### Define grid overlap functions
###   - Python's 'multiprocessing' module seems to cause process lock
###

Lmax = 2

def eval_overlap(grid,P_list, IP,indx):
#    if opts.verbose: 
#        print " Evaluating for ", indx
    global Lmax
    global opts
    P2 = P_list[indx]
#    P2.print_params()
    line_out = []
    line_out = list(grid[indx])
    if not opts.skip_overlap:
        T_here = 1./IP.deltaF
        P2.deltaF=1./T_here
        hf2 = lalsimutils.complex_hoff(P2)
        nm2 = IP.norm(hf2);  hf2.data.data *= 1./nm2
        ip_val = IP.ip(hfBase,hf2)
        line_out.append(ip_val)
    else:
        line_out.append(-1)
    if opts.verbose:
            print(" Answer ", indx, line_out)
    return line_out

def evaluate_overlap_on_grid(hfbase,param_names, grid):
    # Validate grid is working: Create a loop and print for each one.
    # WARNING: Assumes grid for mass-unit variables hass mass units (!)
    global xi_factor
    P_list = []
    for line in grid:
        Pgrid = P.manual_copy()
        # Set attributes that are being changed as necessary, leaving all others fixed
        for indx in np.arange(len(param_names)):
                Pgrid.assign_param(param_names[indx], line[indx])   # DANGER: need to keep overlap values
        # Rescale mass parameters using the xi factor
        xi_now = Pgrid.extract_param('xi')
        eta_now = Pgrid.extract_param('eta')
            # THIS CAN BE NEGATIVE: PATHOLOGICAL
        Pgrid.m1 *= (1.+xi_now*xi_factor - eta_factor0*(0.25-eta_now))
        Pgrid.m2 *= (1.+xi_now*xi_factor - eta_factor0*(0.25-eta_now))
        P_list.append(Pgrid)
#    print "Length check", len(P_list), len(grid)
    ###
    ### Loop over grid and make overlaps : see effective fisher code for wrappers
    ###
    #  FIXME: More robust multiprocessing implementation -- very heavy!
#    p=Pool(n_threads)
    # PROBLEM: Pool code doesn't work in new configuration.
    grid_out = np.array(list(map(functools.partial(eval_overlap, grid, P_list,IP), np.arange(len(grid)))))
    # Remove mass units at end
    for p in ['mc', 'm1', 'm2', 'mtot']:
        if p in param_names:
            indx = param_names.index(p)
            grid_out[:,indx] /= lal.MSUN_SI
    # Truncate grid so overlap with the base point is > opts.min_match. Make sure to CONSISTENTLY truncate all lists (e.g., the P_list)
    grid_out_new = []
    P_list_out_new = []
    for indx in np.arange(len(grid_out)):
      if P_list[indx].m1 >0 and P_list[indx].m2>0:   # reject insane cases where the masses were scaled to be <0
        if (opts.skip_overlap) or (grid_out[indx,-1] > opts.match_value):
            grid_out_new.append(grid_out[indx])
            P_list_out_new.append(P_list[indx])

    grid_out = np.array(grid_out_new)
    return grid_out, P_list_out_new



###
### Define base point 
###


# Handle PSD
# FIXME: Change to getattr call, instead of 'eval'
eff_fisher_psd = lalsim.SimNoisePSDiLIGOSRD
if not opts.psd_file:
    #eff_fisher_psd = eval(opts.fisher_psd)
    eff_fisher_psd = getattr(lalsim, opts.fisher_psd)   # --fisher-psd SimNoisePSDaLIGOZeroDetHighPower   now
    analyticPSD_Q=True
else:
    eff_fisher_psd = lalsimutils.load_resample_and_clean_psd(opts.psd_file, 'H1', 1./opts.seglen)
    analyticPSD_Q = False



P=lalsimutils.ChooseWaveformParams()
if opts.inj:
    from igwn_ligolw import lsctables, table, utils # check all are needed
    filename = opts.inj
    event = opts.event_id
    xmldoc = utils.load_filename(filename, verbose = True,contenthandler =lalsimutils.cthdler)
    sim_inspiral_table = table.get_table(xmldoc, lsctables.SimInspiralTable.tableName)
    P.copy_sim_inspiral(sim_inspiral_table[int(event)])
else:    
    P.m1 = opts.mass1 *lal.MSUN_SI
    P.m2 = opts.mass2 *lal.MSUN_SI
    P.s1z = opts.s1z
    P.dist = 150*1e6*lal.PC_SI
    if opts.eff_lambda and Psig:
        lambda1, lambda2 = 0, 0
        if opts.eff_lambda is not None:
            lambda1, lambda2 = lalsimutils.tidal_lambda_from_tilde(m1, m2, opts.eff_lambda, opts.deff_lambda or 0)
            Psig.lambda1 = lambda1
            Psig.lambda2 = lambda2

    P.fmin=opts.fmin   # Just for comparison!  Obviously only good for iLIGO
    P.ampO=-1  # include 'full physics'
    if opts.approx:
        P.approx = lalsim.GetApproximantFromString(opts.approx)
        if not (P.approx in [lalsim.TaylorT1,lalsim.TaylorT2, lalsim.TaylorT3, lalsim.TaylorT4]):
            # Do not use tidal parameters in approximant which does not implement them
            print(" Do not use tidal parameters in approximant which does not implement them ")
            P.lambda1 = 0
            P.lambda2 = 0
    else:
        P.approx = lalsim.GetApproximantFromString("TaylorT4")
P.deltaT=1./16384
P.taper = lalsim.SIM_INSPIRAL_TAPER_START
P.deltaF = 1./opts.seglen #lalsimutils.findDeltaF(P)
P.fref = opts.fref
P.print_params()


print("    -------INTERFACE ------")
hfBase =None
IP=None
if not opts.skip_overlap:
    hfBase = lalsimutils.complex_hoff(P)
    IP = lalsimutils.CreateCompatibleComplexOverlap(hfBase,analyticPSD_Q=analyticPSD_Q,psd=eff_fisher_psd,fMax=opts.fmax)
    nmBase = IP.norm(hfBase)
    hfBase.data.data *= 1./nmBase
    if opts.verbose:
        print(" ------  SIGNAL DURATION ----- ")
        print(hfBase.data.length*P.deltaT)



###
### If we are testing an existing grid to see if simulations are missing
###

# Step 1: Load in the grid, identify NR simulations

P_list_nr = []
gp_list = []
if opts.inj and opts.insert_missing_spokes:
    import spokes
    sdHere = spokes.LoadSpokeXML(opts.inj)
    for key in sdHere:
        P_here =sdHere[key][0]
        P_list_nr.append(P_here) 
        compare_dict = {}
        compare_dict['q'] = P_here.m2/P_here.m1 # Need to match the template parameter. NOTE: VERY IMPORTANT that P is updated with the event params
        compare_dict['s1z'] = P_here.s1z
        compare_dict['s1x'] = P_here.s1x
        compare_dict['s1y'] = P_here.s1y
        compare_dict['s2z'] = P_here.s2z
        compare_dict['s2x'] = P_here.s2x
        compare_dict['s2y'] = P_here.s2y
        good_sim_list = nrwf.NRSimulationLookup(compare_dict,valid_groups=[opts.group])
        if len(good_sim_list)< 1:
                        print(" ------- NO MATCHING SIMULATIONS FOUND ----- ")
                        import sys
                        sys.exit(0)
                        print(" Identified set of matching NR simulations ", good_sim_list)
        try:
                        print("   Attempting to pick longest simulation matching  the simulation  ")
                        MOmega0  = 1
                        good_sim = None
                        for key in good_sim_list:
                                print(key, nrwf.internal_EstimatePeakL2M2Emission[key[0]][key[1]])
                                if nrwf.internal_WaveformMetadata[key[0]][key[1]]['Momega0'] < MOmega0:
                                        good_sim = key
                                        MOmega0 = nrwf.internal_WaveformMetadata[key[0]][key[1]]['Momega0']
                                print(" Picked  ",key,  " with MOmega0 ", MOmega0, " and peak duration ", nrwf.internal_EstimatePeakL2M2Emission[key[0]][key[1]])
        except:
            good_sim  = good_sim_list[0] # pick the first one.  Note we will want to reduce /downselect the lookup process
            group = good_sim[0]
            param = good_sim[1] 
            gp_list.append( (group,param))



###
### Load in the NR simulation array metadata
###
P_list_NR = []
omega_list_NR=[]

glist = []
if opts.group:
    glist = [opts.group]
else:
    glist = nrwf.internal_ParametersAvailable.keys()

eta_range = eval(opts.eta_range)

print(opts.group, opts.param)
for group in glist:
  if not opts.param:
    for param in nrwf.internal_ParametersAvailable[group]:
     print("  -> ", group, param)
     try:
        wfP = nrwf.WaveformModeCatalog(group,param,metadata_only=True)
        wfP.P.deltaT = P.deltaT
        wfP.P.deltaF = P.deltaF
        wfP.P.fmin = P.fmin
#        wfP.P.print_params()
        # Add parameters. Because we will compare with SEOB, we need an ALIGNED waveform, so we fake it
        if wfP.P.extract_param('eta') >= eta_range[0] and wfP.P.extract_param('eta')<=eta_range[1] and (not np.isnan(wfP.P.s1z) and not np.isnan(wfP.P.s2z)):
            if (not opts.skip_overlap) and wfP.P.SoftAlignedQ():
                print(" Adding aligned sim ", group, param,  " and changing spin values, so we can compute overlaps ")
                wfP.P.approx = lalsim.GetApproximantFromString(opts.approx)  # Make approx consistent and sane
                wfP.P.m2 *= 1. - 1e-6  # Prevent failure for exactly equal!
            # Satisfy error checking condition for lal
                wfP.P.s1x = 0
                wfP.P.s2x = 0
                wfP.P.s1y = 0
                wfP.P.s2y = 0
                P_list_NR = P_list_NR + [wfP.P]
                try:
                    omega_list_NR += [nrwf.internal_WaveformMetadata[group][param]["Momega0"]]
                except:
                    omega_list_NR+=[-1]
            elif opts.skip_overlap:
                if opts.nospin_only:
                    if wfP.P.s1x ==0.0 and wfP.P.s1y==0.0 and wfP.P.s1z==0.0 and wfP.P.s2x==0.0 and wfP.P.s2y == 0.0 and wfP.P.s2z == 0.0 and not opts.eccentricity:
                        print(" Adding non-spinning simulation ", group, param)
                        P_list_NR = P_list_NR + [wfP.P]
                        try:
                            omega_list_NR += [nrwf.internal_WaveformMetadata[group][param]["Momega0"]]
                        except:
                            omega_list_NR+=[-1]
                    elif wfP.P.s1x ==0.0 and wfP.P.s1y==0.0 and wfP.P.s1z==0.0 and wfP.P.s2x==0.0 and wfP.P.s2y == 0.0 and wfP.P.s2z == 0.0 and opts.eccentricity:
                        wfP.P.print_params()
                        if not wfP.P.eccentricity==0:
                            P_list_NR = P_list_NR + [wfP.P]
                            try:
                                omega_list_NR += [nrwf.internal_WaveformMetadata[group][param]["Momega0"]]
                            except:
                                omega_list_NR+=[-1]
                elif not opts.aligned_only:
                    print(" Adding generic sim; for layout only ", group, param)
                    P_list_NR = P_list_NR + [wfP.P]
                    try:
                        omega_list_NR += [nrwf.internal_WaveformMetadata[group][param]["Momega0"]]
                    except:
                        omega_list_NR+=[-1]
                elif opts.aligned_only and  wfP.P.SoftAlignedQ():
                    print(" Adding aligned spin simulation; for layout only", group, param, " and fixing transverse spins accordingly")
                    wfP.P.s1x = 0
                    wfP.P.s2x = 0
                    wfP.P.s1y = 0
                    wfP.P.s2y = 0
                    P_list_NR = P_list_NR + [wfP.P]
                    try:
                        omega_list_NR += [nrwf.internal_WaveformMetadata[group][param]["Momega0"]]
                    except:
                        omega_list_NR+=[-1]
            else:
                print(" Skipping non-aligned simulation because overlaps active (=SEOBNRv2 comparison usually)", group, param)
#                wfP.P.print_params()
#                print nrwf.internal_WaveformMetadata[group][param]
     except:
        print(" Failed to add ", group, param)
        
  else: # target case if a single group and parameter sequence are specified
        print("Looping over list ", opts.param)
        for paramKey in opts.param:
            print(" Adding specific simulation ", opts.group, paramKey)
            if nrwf.internal_ParametersAreExpressions[group]:
                param = eval(paramKey)
            else:
                param = paramKey
        wfP = nrwf.WaveformModeCatalog(group,param,metadata_only=True)
        wfP.P.deltaT = P.deltaT
        wfP.P.deltaF = P.deltaF
        wfP.P.fmin = P.fmin
        P_list_NR = P_list_NR + [wfP.P]
        try:
            omega_list_NR += [nrwf.internal_WaveformMetadata[group][param]["Momega0"]]
        except:
            omega_list_NR+=[-1]


if len(P_list_NR)<1:
    print(" No simulations")
    sys.exit(0)


###
### Define parameter ranges to be changed
###

template_min_freq = opts.fmin
ip_min_freq = opts.fmin



###
### Lay out grid, currently CARTESIAN.   OPTIONAL: Load grid from file
###


# For now, we just extrude in these parameters
if not opts.skip_overlap:  # aligned only!  Compare to SEOB
    if opts.mc_range:
        param_names = ['mc', 'eta','s1z','s2z']
    else:
        param_names = ['mtot', 'q','s1z','s2z']
else:
    if opts.mc_range:
        param_names = ['mc', 'eta', 's1x', 's1y', 's1z', 's2x', 's2y', 's2z']
    else:
        param_names = ['mtot', 'q', 's1x', 's1y', 's1z', 's2x', 's2y', 's2z']
    if opts.eccentricity:
        param_names.append('eccentricity')

mass_range =[]
if opts.mc_range:
    mass_range = np.array(eval(opts.mc_range))*lal.MSUN_SI
else:
    mass_range = np.array(eval(opts.mtot_range))*lal.MSUN_SI
mass_grid =np.linspace( mass_range[0],mass_range[1],int(opts.grid_cartesian_npts))

# Loop over simulations and mass grid
grid = []
indx=-1
print("Length check", len(P_list_NR), len(omega_list_NR))
for P in P_list_NR:
    indx+=1
    Momega0_here = omega_list_NR[indx]
#    P.print_params()
    eta0 = P.extract_param('eta')   # prevent drift in mass ratio across the set if I change mc
    for M in mass_grid:
        if opts.mc_range:
            P.assign_param('mc', M)
            P.assign_param('eta', eta0)  # prevent drift in mass ratio
        else:
            P.assign_param('mtot', M)
        newline = []
        for param in param_names:
            newline = newline + [P.extract_param(param)]
#        print newline
        fstart_NR = Momega0_here/((P.extract_param('mtot')/(lal.MSUN_SI) ) *lalsimutils.MsunInSec*(np.pi))
        if   fstart_NR < P.fmin or not opts.require_fmin_above_NR_start :
            # only add mass point if a valid mass choice, given NR starting frequency.
            grid = grid+[ newline]

print(" ---- DONE WITH GRID SETUP --- ")
print(" grid points # = " ,len(grid))


grid_out, P_list = evaluate_overlap_on_grid(hfBase, param_names, grid)
if len(grid_out)==0:
    print(" No points survive....")

###
### Optional: Write grid to XML file (ONLY if using cutoff option)
###
lalsimutils.ChooseWaveformParams_array_to_xml(P_list, fname=opts.fname, fref=P.fref)


###
### Write output to text file:  p1 p2 p3 ... overlap, only including named params
###
headline = ' '.join(param_names + ['ip'])
if int(np.version.short_version.split('.')[1])>=7:  # http://docs.scipy.org/doc/numpy/reference/generated/numpy.savetxt.html
    np.savetxt(opts.fname+".dat", grid_out, header=headline)
else:
    np.savetxt(opts.fname+".dat", grid_out)   # 



###
### Optional: Scatterplot
###
if opts.verbose and len(param_names)==1 and len(grid_out)>0:
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.plot(grid_out[:,0], grid_out[:,1])
    plt.show()

if opts.verbose and len(param_names)==2:
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(grid_out[:,0], grid_out[:,1], grid_out[:,2])
    plt.show()

print(" ---- DONE ----")



# -*- coding: utf-8 -*-
'''
Monte Carlo Integrator
----------------------
Perform an adaptive monte carlo integral.
'''
from __future__ import print_function
import numpy as np
from . import gaussian_mixture_model as GMM
import traceback
import time
from scipy.special import logsumexp

regularize_log_scale = 1e-64  # before taking np.log, add this, so we don't propagate infinities

try:
    from multiprocess import Pool
except:
    print('no multiprocess')


class integrator:
    '''
    Class to iteratively perform an adaptive Monte Carlo integral where the integrand
    is a combination of one or more Gaussian curves, in one or more dimensions.

    Parameters
    ----------
    d : int
        Total number of dimensions.

    bounds : dictionary with array bounds, with keys matching gmm_dict
        Limits of integration, where each row represents [left_lim, right_lim]
        for its corresponding dimension.

    gmm_dict : dict
        Dictionary where each key is a tuple of one or more dimensions
        that are to be modeled together. If the integrand has strong correlations between
        two or more dimensions, they should be grouped. Each value is by default initialized
        to None, and is replaced with the GMM object for its dimension(s).

    n_comp : int or {tuple:int}
        The number of Gaussian components per group of dimensions. If its type is int,
        this number of components is used for all dimensions. If it is a dict, it maps
        each key in gmm_dict to an integer number of mixture model components.

    n : int
        Number of samples per iteration

    prior : function
        Function to evaluate prior for samples

    user_func : function
        Function to run each iteration

    L_cutoff : float
        Likelihood cutoff for samples to store

    use_lnL : bool
        Whether or not lnL or L will be returned by the integrand
    '''

    def __init__(self, d, bounds, gmm_dict, n_comp, n=None, prior=None,
                user_func=None, proc_count=None, L_cutoff=None, use_lnL=False,return_lnI=False,gmm_adapt=None,gmm_epsilon=None,tempering_exp=1,temper_log=False,lnw_failure_cut=None):
        # if 'return_lnI' is active, 'integral' holds the *logarithm* of the integral.
        # user-specified parameters
        self.d = d
        self.bounds = bounds
        self.gmm_dict = gmm_dict
        self.gmm_adapt = gmm_adapt
        self.gmm_epsilon= gmm_epsilon
        self.n_comp = n_comp
        self.user_func=user_func
        self.prior = prior
        self.proc_count = proc_count
        self.use_lnL = use_lnL
        self.return_lnI = return_lnI
        # constants
        self.t = 0.02 # percent estimated error threshold
        if n is None:
            self.n = int(5000 * self.d) # number of samples per batch
        else:
            self.n = int(n)
        self.ntotal = 0
        # integrator object parameters
        self.sample_array = None
        self.value_array = None
        self.sampling_prior_array = None
        self.prior_array = None
        self.scaled_error_squared = 0.
        if self.return_lnI:
            self.scaled_error_squared = None
        if not(lnw_failure_cut):
            self.terrible_lnw_threshold = -1000
        else:
            self.terrible_lnw_threshold = lnw_failure_cut
        self.log_error_scale_factor = 0.
        self.integral = 0
        if self.return_lnI:
            self.integral=None
        self.eff_samp = 0
        self.iterations = 0 # for weighted averages and count
        self.log_scale_factor = 0  # to handle very large answers
        self.max_value = float('-inf') # for calculating eff_samp
        self.total_value = 0 # for calculating eff_samp
        if self.return_lnI:
            self.total_value = None
        self.n_max = float('inf')
        # saved values
        self.cumulative_samples = np.empty((0, d))
        self.cumulative_values = np.empty(0)
        self.cumulative_p = np.empty(0)
        self.cumulative_p_s = np.empty(0)
        self.tempering_exp=tempering_exp
        self.temper_log=temper_log
        if L_cutoff is None:
            self.L_cutoff = -1
        else:
            self.L_cutoff = L_cutoff
        
    def _calculate_prior(self):
        if self.prior is None:
            self.prior_array = np.ones(self.n)
        else:
            self.prior_array = self.prior(self.sample_array).flatten()

    def _sample(self):
        self.sampling_prior_array = np.ones(self.n)
        self.sample_array = np.empty((self.n, self.d))
        for dim_group in self.gmm_dict: # iterate over grouped dimensions
            # create a matrix of the left and right limits for this set of dimensions
            new_bounds = np.empty((len(dim_group), 2))
            new_bounds = self.bounds[dim_group]
            if len(new_bounds.shape) < 2:
                new_bounds = np.array([new_bounds])
            # index = 0
            # for dim in dim_group:
            #     new_bounds[index] = self.bounds[dim]
            #     index += 1
            model = self.gmm_dict[dim_group]
            if model is None:
                # sample uniformly for this group of dimensions
                llim = new_bounds[:,0]
                rlim = new_bounds[:,1]
                temp_samples = np.random.uniform(llim, rlim, (self.n, len(dim_group)))
                # update responsibilities
                vol = np.prod(rlim - llim)
                self.sampling_prior_array *= 1.0 / vol
            else:
                # sample from the gmm
                temp_samples = model.sample(self.n)#, new_bounds)
                # update responsibilities
                self.sampling_prior_array *= model.score(temp_samples)#, new_bounds)
            index = 0
            for dim in dim_group:
                # put columns of temp_samples in final places in sample_array
                self.sample_array[:,dim] = temp_samples[:,index]
                index += 1

    def _train(self):
        sample_array, value_array, sampling_prior_array = np.copy(self.sample_array), np.copy(self.value_array), np.copy(self.sampling_prior_array)
        if self.use_lnL:
            lnL = value_array
        else:
            lnL = np.log(value_array+regularize_log_scale) # note we can get negative infinity here
        log_weights = self.tempering_exp*lnL + np.log(self.prior_array) - sampling_prior_array
        if self.temper_log:
            log_weights =np.log(np.maximum(lnL,1e-5))   # simplest to do it this way
        for dim_group in self.gmm_dict: # iterate over grouped dimensions
            if self.gmm_adapt:
                if (dim_group in self.gmm_adapt):
                    if not(self.gmm_adapt[dim_group]):   # disabling adaptation requires user *specifically request* not to use that dimension set; all other choices lead to adaptation
                        continue
            # create a matrix of the left and right limits for this set of dimensions
            new_bounds = np.empty((len(dim_group), 2))
            new_bounds = self.bounds[dim_group]
#            index = 0
#            for dim in dim_group:
#                new_bounds[index] = self.bounds[dim]
#                index += 1
            model = self.gmm_dict[dim_group] # get model for this set of dimensions
            temp_samples = np.empty((self.n, len(dim_group)))
            index = 0
            for dim in dim_group:
                # get samples corresponding to the current model
                temp_samples[:,index] = sample_array[:,dim]
                index += 1
            if model is None:
                # model doesn't exist yet
                if isinstance(self.n_comp, int) and self.n_comp != 0:
                    model = GMM.gmm(self.n_comp, new_bounds,epsilon=self.gmm_epsilon)
                    model.fit(temp_samples, log_sample_weights=log_weights)
                elif isinstance(self.n_comp, dict) and self.n_comp[dim_group] != 0:
                    model = GMM.gmm(self.n_comp[dim_group], new_bounds,epsilon=self.gmm_epsilon)
                    model.fit(temp_samples, log_sample_weights=log_weights)
            else:
                model.update(temp_samples, log_sample_weights=log_weights)
            try:
                # Verify model can evaluated! Quick and dirty test to confirm not singular
                model.score(temp_samples[:5])
                self.gmm_dict[dim_group] = model
            except:
                print(" Failed to update ", dim_group)


    def _calculate_results(self):
        if self.use_lnL:
            lnL = np.copy(self.value_array) # changing the naming convention, just for this function, now that I know better
        else:
            lnL = np.log(self.value_array+regularize_log_scale)
        # strip off any samples with likelihoods less than our cutoff
        mask = np.ones(lnL.shape,dtype=bool)
        if not(self.L_cutoff is None):  # if not none
            if not(np.isinf(self.L_cutoff)):  # and not infinite, then apply the cutoff
                mask = lnL > (np.log(self.L_cutoff) if self.L_cutoff > 0 else -np.inf)
#        print(mask, self.L_cutoff, lnL)
        lnL = lnL[mask]
        prior = self.prior_array[mask]
        sampling_prior = self.sampling_prior_array[mask]
        
        # append to the cumulative arrays
        self.cumulative_samples = np.append(self.cumulative_samples, self.sample_array[mask], axis=0)
        self.cumulative_values = np.append(self.cumulative_values, lnL, axis=0)
        self.cumulative_p = np.append(self.cumulative_p, prior, axis=0)
        self.cumulative_p_s = np.append(self.cumulative_p_s, sampling_prior, axis=0)
        
        # compute the log sample weights
        log_weights = lnL + np.log(prior) - np.log(sampling_prior)
        if np.any(np.isnan(log_weights)):
            print(" NAN weight ")
            raise ValueError
        if self.terrible_lnw_threshold: 
            if np.max(log_weights) <self.terrible_lnw_threshold:  #this is a terrible fit. Note the ABSOLUTE SCALE, which is problem dependent!
                print(" TERRIBLE FIT ")
                raise ValueError

        # do a shift so that the highest log weight is 0, keeping track of the shift
        log_scale_factor = np.max(log_weights) # don't insert nan here
        if not(self.return_lnI):
            scale_factor = np.exp(log_scale_factor)
            log_weights -= log_scale_factor
            summed_vals = scale_factor * np.sum(np.exp(log_weights))
            integral_value = summed_vals / self.n
        
            # Calculate the log of the Monte Carlo integration error.
            # Let `a` be the scale factor, and let `w` be the weights with the scale factor divided out (i.e. weight = a w), then
            #     error^2 = var(a w) / N = a^2 var(w) / N.
            # The point is that 0 <= w <= 1, so there's no potential for overflow here when calculating the variance.
            # Since the scale factor `a` is potentially very large, squaring it could overflow; therefore, we keep out a factor of ln(a^2) = 2ln(a).
            # Note that ln(a) is exactly the log_scale_factor variable defined above.
            scaled_error_squared = np.var(np.exp(log_weights)) / self.n
            log_error_scale_factor = 2. * log_scale_factor
        
            self.integral = (self.iterations * self.integral + integral_value) / (self.iterations + 1)
        
            # Calculate the running average of the variance.
            # This calculation is complicated by the scale factors: the previous running average is scaled by exp(self.log_error_scale_factor),
            # which is currently the log_error_scale_factor from the previous iteration.
            # We can avoid overflows by factoring our *new* log_error_scale_factor out of the running average, so that we exponentiate a smaller number.
            self.scaled_error_squared = (self.iterations * np.exp(self.log_error_scale_factor - log_error_scale_factor) * self.scaled_error_squared + scaled_error_squared) / (self.iterations + 1)
            self.log_error_scale_factor = log_error_scale_factor # having computed the running average, update this
        
            # calculate the effective samples
            self.total_value += summed_vals
            self.max_value = max(scale_factor, self.max_value)
            self.eff_samp = self.total_value / self.max_value
        else: # using lnI return values
            # Evaluate integral and effective sample count
            # Note with lnI return values, no more need to keep track of a scale factor at all
            log_sum_weights = logsumexp(log_weights) 
            log_integral_here = log_sum_weights - np.log(self.n)
            if not(self.integral ):
                self.integral = log_integral_here  
                self.total_value = log_sum_weights  
                self.max_value = log_scale_factor
            else:
                self.integral = logsumexp([ self.integral +np.log(self.iterations), log_integral_here]) - np.log(self.iterations+1)
                self.total_value= logsumexp([self.total_value, log_sum_weights])
                self.max_value = np.max([self.max_value,np.max(log_weights)])
            self.eff_samp = np.exp(self.total_value - (self.max_value  ))

            # Evaluate error squared, avoiding overflow
            tmp_max = np.max(log_weights)
            log_scaled_error_squared = np.log(np.var(np.exp(log_weights - tmp_max))) + 2*tmp_max   - np.log(self.n)
            if not(self.scaled_error_squared):
                self.scaled_error_squared = log_scaled_error_squared
            else:
                self.scaled_error_squared = logsumexp([ self.scaled_error_squared + np.log(self.iterations), log_scaled_error_squared]) - np.log(self.iterations+1)

    def _reset(self):
        ### reset GMMs
        for k in self.gmm_dict:
            self.gmm_dict[k] = None
        

    def integrate(self, func, min_iter=10, max_iter=20, var_thresh=0.0, max_err=10,
            neff=float('inf'), nmax=None, progress=False, epoch=None,verbose=True,force_no_adapt=False,use_lnL=False,return_lnI=False,**kwargs):
        '''
        Evaluate the integral

        Parameters
        ----------
        func : function
            Integrand function
        min_iter : int
            Minimum number of integrator iterations
        max_iter : int
            Maximum number of integrator iterations
        var_thresh : float
            Variance threshold for terminating integration
        max_err : int
            Maximum number of errors to catch before terminating integration
        neff : float
            Effective samples threshold for terminating integration
        nmax : int
            Maximum number of samples to draw
        progress : bool
            Print GMM parameters each iteration
        n_adapt: number of *adaptations* we will perform, before freezing the GMM
        '''
        n_adapt = int(kwargs["n_adapt"]) if "n_adapt" in kwargs else 100
        tripwire_fraction = kwargs["tripwire_fraction"] if "tripwire_fraction" in kwargs else 2  # make it impossible to trigger
        tripwire_epsilon = kwargs["tripwire_epsilon"] if "tripwire_epsilon" in kwargs else 0.001 # if we are not reasonably far away from unity, fail!
        self.use_lnL = use_lnL   # option to override initialization of sampler : argument passed in another way
        self.return_lnI = return_lnI   # option to override initialization of sampler : argument passed in another way


        err_count = 0
        cumulative_eval_time = 0
        adapting=True
        if nmax is None:
            nmax = max_iter * self.n
        while self.iterations < max_iter and self.ntotal < nmax and self.eff_samp < neff:
            if (self.ntotal > nmax*tripwire_fraction) and (self.eff_samp < 1+tripwire_epsilon):
                print(" Tripwire: n_eff too low ")
                raise Exception("Tripwire on n_eff")

            if force_no_adapt or self.iterations >= n_adapt:
                adapting=False
#            print('Iteration:', self.iterations)
            if err_count >= max_err:
                print('Exiting due to errors...')
                break
            try:
                self._sample()
            except KeyboardInterrupt:
                print('KeyboardInterrupt, exiting...')
                break
            except Exception as e:
                print(traceback.format_exc())
                print('Error sampling, resetting...')
                err_count += 1
                self._reset()
                continue
            t1 = time.time()
            if self.proc_count is None:
                self.value_array = func(np.copy(self.sample_array)).flatten()
            else:
                split_samples = np.array_split(self.sample_array, self.proc_count)
                p = Pool(self.proc_count)
                self.value_array = np.concatenate(p.map(func, split_samples), axis=0)
                p.close()
            cumulative_eval_time += time.time() - t1
            self._calculate_prior()
            try:
                self._calculate_results()
            except KeyboardInterrupt:
                print('KeyboardInterrupt, exiting...')
                break
            except Exception as e:
                print(traceback.format_exc())
                print('Error calculating results, resetting...')
                err_count += 1
                self._reset()
                continue
            self.iterations += 1
            self.ntotal += self.n
            testval =self.scaled_error_squared
            if not(self.return_lnI):
                testval = np.log(self.scaled_error_squared) + self.log_error_scale_factor
            if self.iterations >= min_iter and testval < np.log(var_thresh):
                break
            try:
                if adapting:
                    self._train()
            except KeyboardInterrupt:
                print('KeyboardInterrupt, exiting...')
                break
            except Exception as e:
                print(traceback.format_exc())
                print('Error training, resetting...')
                err_count += 1
                self._reset()
            if self.user_func is not None:
                self.user_func(self)
            if progress:
                for k in self.gmm_dict:
                    if self.gmm_dict[k] is not None:
                        self.gmm_dict[k].print_params()
            if epoch is not None and self.iterations % epoch == 0:
                self._reset()
            if verbose:
                # Standard mcsampler message, to monitor convergence
                if not(self.return_lnI):
                    print(" : {} {} {} {} {} ".format((self.iterations-1)*self.n, self.eff_samp, np.sqrt(2*np.max(self.cumulative_values)), np.sqrt(2*(np.log(self.integral))),  np.sqrt(self.scaled_error_squared )/self.integral/np.sqrt(self.iterations ) ) )
                else:
                    print(" : {} {} {} {} {} ".format((self.iterations-1)*self.n, self.eff_samp, np.sqrt(2*np.max(self.cumulative_values)), np.sqrt(2*self.integral), np.exp(0.5*(self.scaled_error_squared - self.integral*2) )/np.sqrt(self.iterations)))
        print('cumulative eval time: ', cumulative_eval_time)
        print('integrator iterations: ', self.iterations)



#! /bin/bash
# construct_MDC_pipeline_wrapper <event_id> <approximant>
#     Creates a new directory with the event ID
#     Generates a dag (and a 'test command') for processing that event
#
# ASSUMES (for now)
#     - Just H1, L1 (!).  Needs to implement variable IFO list!
#     - cache file 'local.cache' exists in current working directory, holding needed locations of files
#     - 1_injections.xml file exists in current working directory, holding all injections
#     - H1_psd.xml.gz and L1_psd.xml.gz exist in the current working directory
#     - channel-name file

export ILE=`which integrate_likelihood_extrinsic`
export CME=`which compute_marginalized_likelihood`
export EWD=`which util_EstimateWaveformDuration.py`
echo Using ILE ${ILE} 
echo Using CME ${CME} 
echo Using EWD ${EWD}
mkdir -p $1
cd $1

export event=$1
if [ -n "$2" ]; then
  export approximant=$2
else
  export approximant=TaylorT4
fi
echo " Constructing directory for ", $event, " using approximant ", $approximant

cp ../1_injections.xml inj.xml
cp ../local.cache local.cache
cp ../H1_psd.xml.gz .
cp ../L1_psd.xml.gz .
cp ../channel-name .

export MASS1=`util_PrintInjectionParameters --inj inj.xml --event $1 --indicate-mass1 1 | tail -1`
export MASS2=`util_PrintInjectionParameters --inj inj.xml --event $1 --indicate-mass2 1 | tail -1`
export EVENT_TIME=`util_PrintInjectionParameters --inj inj.xml --event $1 --signal-time 0.0001  | tail -1`
export BETA=`util_AdaptiveExponent.py --Niter 4000 --inj inj.xml --event $1 --channel-name H1= --psd-file-singleifo "H1=H1_psd.xml.gz" --channel-name L1= --psd-file-singleifo "L1=L1_psd.xml.gz" | tail -1`



# CHANNEL NAMES
#    - by default, use the ER channel names -- that is what I will use for now
INJ_CHANNEL_NAME=`cat channel-name`
GDB_V_INJ_CHANNEL_NAME=FAKE_h_16384Hz_4R


## Make DAG

# Write a command that has the same options but just evaluates the likliehood (marginalized) at the trigger masses.  Helpful for debugging setup
#    - Comments left to indicate kind of arguments ROS wants on master
echo  ${ILE}  --cache-file local.cache   --event-time ${EVENT_TIME} --mass1 ${MASS1} --mass2 ${MASS2} --channel-name H1=${INJ_CHANNEL_NAME} --channel-name L1=${INJ_CHANNEL_NAME}  --psd-file "H1=H1_psd.xml.gz" --psd-file "L1=L1_psd.xml.gz"  --reference-freq 0 --save-samples    --time-marginalization --n-max 200000 --n-eff 1000 --output-file ile-mdc-${event}.xml.gz   --save-P 0.0001 --fmax 2000 --adapt-weight-exponent ${BETA} --adapt-floor-level 0.1 --n-chunk 4000 --approximant $approximant --convergence-tests-on   > testme-command.sh  #  --adapt-parameter right_ascension --adapt-parameter declination --adapt-parameter distance       
# add postprocessing
echo make_triplot ile-mdc-${event}.xml.gz -o ${event}-triplot.pdf >> testme-command.sh
echo plot_integral ile-mdc-${event}.xml.gz --output integral.pdf >> testme-command.sh
chmod a+x testme-command.sh

# Make another command, to test operation at a known sky location.  Useful comparison against all-sky search
echo  ${ILE}  --cache-file local.cache   --event-time ${EVENT_TIME} --mass1 ${MASS1} --mass2 ${MASS2} `xmlInspiralToILEPinned.py inj.xml ${event} right_ascension declination`  --channel-name H1=${INJ_CHANNEL_NAME} --channel-name L1=${INJ_CHANNEL_NAME}  --psd-file "H1=H1_psd.xml.gz" --psd-file "L1=L1_psd.xml.gz"  --reference-freq 0 --save-samples    --time-marginalization --n-max 200000 --n-eff 1000 --output-file ile-mdc-${event}.xml.gz   --save-P 0.0001 --fmax 2000 --adapt-weight-exponent ${BETA} --adapt-floor-level 0.1 --n-chunk 4000 --approximant $approximant --convergence-tests-on   > testme-command-skypinned.sh  #  --adapt-parameter right_ascension --adapt-parameter declination --adapt-parameter distance       
# add postprocessing
echo make_triplot ile-mdc-${event}.xml.gz -o ${event}-triplot.pdf >> testme-command-skypinned.sh
echo plot_integral ile-mdc-${event}.xml.gz --output integral.pdf >> testme-command-skypinned.sh
chmod a+x testme-command-skypinned.sh




# Write the actual DAG
#   - large n-max chosen for prototyping purposes.  Hopefully we will hit the n-eff limit before reaching it.
${CME} --cache-file local.cache   --event-time ${EVENT_TIME} --mass1 ${MASS1} --mass2 ${MASS2} --channel-name H1=${INJ_CHANNEL_NAME} --channel-name L1=${INJ_CHANNEL_NAME} --psd-file "H1=H1_psd.xml.gz" --psd-file "L1=L1_psd.xml.gz"    --save-samples  --time-marginalization --n-max 1000000 --n-eff 1000 --output-file CME-${event}.xml.gz   --save-P 0.0001  --n-copies 10 --fmax 2000 --adapt-weight-exponent ${BETA} --adapt-floor-level 0.1 --n-chunk 4000  --approximant $approximant --convergence-tests-on  # --adapt-parameter right_ascension --adapt-parameter declination --adapt-parameter distance  

# Write a command to convert the result to a flat ascii grid in m1,m2, lnL.  Ideally part of postprocessing DAG
echo 'for i in CME-*.xml.gz; do ligolw_print -t sngl_inspiral -c mass1 -c mass2 -c snr -c tau0  -d ' ' $i; done > massgrid.txt &' > postprocess-massgrid.sh
cat >> postprocess-massgrid.sh <<EOF
convert_output_format_ile2inference  CME-*.xml.gz > flatfile-points.dat
postprocess_1d_cumulative --save-sampler-file flatfile --inj inj.xml --event $event
ln -s flatfile-points.dat flatfile_noweight-points.dat
postprocess_1d_cumulative --adapt-beta 0 --save-sampler-file flatfile_noweight --disable-1d-density   # to see where skymap samples go
gzip flatfile-points.dat
cat ILE_MASS*.cache > net-ile.cache 
ligolw_sqlite CME-*.xml.gz -d net-ile.sqlite
extract_one_injection inj.xml $event
#make_triplot net-ile.sqlite --injection  selected_event.xml.gz
EOF
chmod a+x postprocess-massgrid.sh

echo "===== TO SUBMIT THE DAG ====== "
echo condor_submit_dag `pwd`/marginalize-extrinsic.dag

echo "===== TO TEST THE UNDERLYING INFRASTRUCTURE USING THIS DATA ====== "
echo  `pwd`/testme-command.sh 



import numpy as np
from matplotlib import pylab as plt

#ILE_fmins = [10., 15., 20., 25., 30., 35., 40.]
ILE_data = [8664., 2172., 2674., 2174., 2162., 1892., 1890.] # runtimes in s (avg of 2 runs)
ILE_fmins = [15., 20., 25., 30., 35., 40.]
T1_data = [2195., 2679., 2124., 2189., 1885., 1861.]
EOB_data = [2202., 2407., 1859., 1847., 1584., 1564.]

rt_40 = 1861. # runtime of Bayesian PE using fmin=40.

def Bayes_runtime(fmin, rt_40):
    """
    Compute runtime of Bayesian PE for a min. freq. fmin
    given the rt_40 as the runtime when fmin=40 Hz.
    """
    return rt_40 * (fmin / 40.)**(-8./3.)

fmins = np.arange(10., 40., 0.1)

plt.semilogy(fmins, Bayes_runtime(fmins, rt_40), 'k-', label="Waveform-limited scaling")
plt.scatter(ILE_fmins, T1_data, label="rapid PE (T1)", c='b')
plt.scatter(ILE_fmins, EOB_data, label="rapid PE (EOBNRv2)", marker='x', c='r')
plt.title('PE Runtimes for $(1.55+1.23) M_\odot$ BNS signal')
plt.legend()
plt.xlabel("min. freq. (Hz)")
plt.ylabel("runtime (s)")
plt.show()



#! /usr/bin/env python
"""
util_MassGrid.py <fnames>
    - prints out <m1> <m2> <lnL>  <neff> for each individual ILE run, to stdout
    - fnames can be an arbitrary mix of .sqlite or .xml files

Used to generate ascii dumps summarizing runs
"""
import sys
import os
import RIFT.misc.xmlutils as xmutils
from optparse import OptionParser
from igwn_ligolw import lsctables, table, utils

# optp= OptionParser()
# opts, args = optp.parse_args()

for fname in sys.argv[1:]:
    fExtension = fname.split(".")[-1]
    # if .sqlite extension
    if fExtension == "sqlite":
        samples = xmlutils.db_to_samples(fname, lsctables.SnglInspiralTable, ['mass1', 'mass2', 'snr', 'tau0', 'event_duration'])
        for row in samples:  # print each individual row, don't reweight yet
            print(row.mass1, row.mass2, row.snr, row.tau0, row.event_duration)
    # otherwise xml: could have just used ligolw_print
    elif fExtension == "xml" or fExtension == "gz":
        samples = table.get_table(utils.load_filename(fname), lsctables.SnglInspiralTable.tableName)
        for row in samples:  # print each individual row, don't reweight yet
            print(row.mass1, row.mass2, row.snr, row.tau0, row.event_duration)

